#!/usr/bin/env python

##
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import, division, print_function

import os
import sys

import avro.datafile
import avro.io
import avro.schema
from avro.codecs import Codecs

try:
    unicode
except NameError:
    unicode = str

NULL_CODEC = 'null'
CODECS_TO_VALIDATE = Codecs.supported_codec_names()

DATUM = {
    'intField': 12,
    'longField': 15234324,
    'stringField': unicode('hey'),
    'boolField': True,
    'floatField': 1234.0,
    'doubleField': -1234.0,
    'bytesField': b'12312adf',
    'nullField': None,
    'arrayField': [5.0, 0.0, 12.0],
    'mapField': {unicode('a'): {'label': unicode('a')},
                 unicode('bee'): {'label': unicode('cee')}},
    'unionField': 12.0,
    'enumField': 'C',
    'fixedField': b'1019181716151413',
    'recordField': {'label': unicode('blah'),
                    'children': [{'label': unicode('inner'), 'children': []}]},
}


def generate(schema_path, output_path):
    with open(schema_path, 'r') as schema_file:
        interop_schema = avro.schema.parse(schema_file.read())
    for codec in CODECS_TO_VALIDATE:
        filename = output_path
        if codec != NULL_CODEC:
            base, ext = os.path.splitext(output_path)
            filename = base + "_" + codec + ext
        with avro.datafile.DataFileWriter(open(filename, 'wb'), avro.io.DatumWriter(),
                                          interop_schema, codec=codec) as dfw:
            dfw.append(DATUM)


if __name__ == "__main__":
    generate(sys.argv[1], sys.argv[2])



#!/usr/bin/env python

##
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
# https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Contains Codecs for Python Avro.

Note that the word "codecs" means "compression/decompression algorithms" in the
Avro world (https://avro.apache.org/docs/current/spec.html#Object+Container+Files),
so don't confuse it with the Python's "codecs", which is a package mainly for
converting charsets (https://docs.python.org/3/library/codecs.html).
"""

from __future__ import absolute_import, division, print_function

import io
import struct
import sys
import zlib
from abc import ABCMeta, abstractmethod
from binascii import crc32
from struct import Struct

import avro.io
from avro.schema import AvroException

#
# Constants
#
STRUCT_CRC32 = Struct('>I')  # big-endian unsigned int


try:
    import bz2
    has_bzip2 = True
except ImportError:
    has_bzip2 = False
try:
    import snappy
    has_snappy = True
except ImportError:
    has_snappy = False
try:
    import zstandard as zstd
    has_zstandard = True
except ImportError:
    has_zstandard = False


class Codec:
    """Abstract base class for all Avro codec classes."""
    __metaclass__ = ABCMeta

    @abstractmethod
    def compress(self, data):
        """Compress the passed data.

        :param data: a byte string to be compressed
        :type data: str

        :rtype: tuple
        :return: compressed data and its length
        """
        pass

    @abstractmethod
    def decompress(self, readers_decoder):
        """Read compressed data via the passed BinaryDecoder and decompress it.

        :param readers_decoder: a BinaryDecoder object currently being used for
                                reading an object container file
        :type readers_decoder: avro.io.BinaryDecoder

        :rtype: avro.io.BinaryDecoder
        :return: a newly instantiated BinaryDecoder object that contains the
                 decompressed data which is wrapped by a StringIO
        """
        pass


class NullCodec(Codec):
    def compress(self, data):
        return data, len(data)

    def decompress(self, readers_decoder):
        readers_decoder.skip_long()
        return readers_decoder


class DeflateCodec(Codec):
    def compress(self, data):
        # The first two characters and last character are zlib
        # wrappers around deflate data.
        compressed_data = zlib.compress(data)[2:-1]
        return compressed_data, len(compressed_data)

    def decompress(self, readers_decoder):
        # Compressed data is stored as (length, data), which
        # corresponds to how the "bytes" type is encoded.
        data = readers_decoder.read_bytes()
        # -15 is the log of the window size; negative indicates
        # "raw" (no zlib headers) decompression.  See zlib.h.
        uncompressed = zlib.decompress(data, -15)
        return avro.io.BinaryDecoder(io.BytesIO(uncompressed))


if has_bzip2:
    class BZip2Codec(Codec):
        def compress(self, data):
            compressed_data = bz2.compress(data)
            return compressed_data, len(compressed_data)

        def decompress(self, readers_decoder):
            length = readers_decoder.read_long()
            data = readers_decoder.read(length)
            uncompressed = bz2.decompress(data)
            return avro.io.BinaryDecoder(io.BytesIO(uncompressed))


if has_snappy:
    class SnappyCodec(Codec):
        def compress(self, data):
            compressed_data = snappy.compress(data)
            # A 4-byte, big-endian CRC32 checksum
            compressed_data += STRUCT_CRC32.pack(crc32(data) & 0xffffffff)
            return compressed_data, len(compressed_data)

        def decompress(self, readers_decoder):
            # Compressed data includes a 4-byte CRC32 checksum
            length = readers_decoder.read_long()
            data = readers_decoder.read(length - 4)
            uncompressed = snappy.decompress(data)
            checksum = readers_decoder.read(4)
            self.check_crc32(uncompressed, checksum)
            return avro.io.BinaryDecoder(io.BytesIO(uncompressed))

        def check_crc32(self, bytes, checksum):
            checksum = STRUCT_CRC32.unpack(checksum)[0]
            if crc32(bytes) & 0xffffffff != checksum:
                raise schema.AvroException("Checksum failure")


if has_zstandard:
    class ZstandardCodec(Codec):
        def compress(self, data):
            compressed_data = zstd.ZstdCompressor().compress(data)
            return compressed_data, len(compressed_data)

        def decompress(self, readers_decoder):
            length = readers_decoder.read_long()
            data = readers_decoder.read(length)
            uncompressed = bytearray()
            dctx = zstd.ZstdDecompressor()
            with dctx.stream_reader(io.BytesIO(data)) as reader:
                while True:
                    chunk = reader.read(16384)
                    if not chunk:
                        break
                    uncompressed.extend(chunk)
            return avro.io.BinaryDecoder(io.BytesIO(uncompressed))


class Codecs(object):
    @staticmethod
    def get_codec(codec_name):
        codec_name = codec_name.lower()
        if codec_name == "null":
            return NullCodec()
        elif codec_name == "deflate":
            return DeflateCodec()
        elif codec_name == "bzip2" and has_bzip2:
            return BZip2Codec()
        elif codec_name == "snappy" and has_snappy:
            return SnappyCodec()
        elif codec_name == "zstandard" and has_zstandard:
            return ZstandardCodec()
        else:
            raise ValueError("Unsupported codec: %r" % codec_name)

    @staticmethod
    def supported_codec_names():
        codec_names = ['null', 'deflate']
        if has_bzip2:
            codec_names.append('bzip2')
        if has_snappy:
            codec_names.append('snappy')
        if has_zstandard:
            codec_names.append('zstandard')
        return codec_names



#!/usr/bin/env python

##
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
# https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Command-line tool

NOTE: The API for the command-line tool is experimental.
"""

from __future__ import absolute_import, division, print_function

import os.path
import sys
import threading
import warnings

import avro.io
from avro import datafile, ipc, protocol

try:
    import BaseHTTPServer as http_server  # type: ignore
except ImportError:
    import http.server as http_server  # type: ignore

try:
    from urllib.parse import urlparse
except ImportError:
    from urlparse import urlparse  # type: ignore


class GenericResponder(ipc.Responder):
    def __init__(self, proto, msg, datum):
        proto_json = open(proto, 'rb').read()
        ipc.Responder.__init__(self, protocol.parse(proto_json))
        self.msg = msg
        self.datum = datum

    def invoke(self, message, request):
        if message.name == self.msg:
            print("Message: %s Datum: %s" % (message.name, self.datum), file=sys.stderr)
            # server will shut down after processing a single Avro request
            global server_should_shutdown
            server_should_shutdown = True
            return self.datum


class GenericHandler(http_server.BaseHTTPRequestHandler):
    def do_POST(self):
        self.responder = responder
        call_request_reader = ipc.FramedReader(self.rfile)
        call_request = call_request_reader.read_framed_message()
        resp_body = self.responder.respond(call_request)
        self.send_response(200)
        self.send_header('Content-Type', 'avro/binary')
        self.end_headers()
        resp_writer = ipc.FramedWriter(self.wfile)
        resp_writer.write_framed_message(resp_body)
        if server_should_shutdown:
            print("Shutting down server.", file=sys.stderr)
            quitter = threading.Thread(target=self.server.shutdown)
            quitter.daemon = True
            quitter.start()


def run_server(uri, proto, msg, datum):
    url_obj = urlparse(uri)
    server_addr = (url_obj.hostname, url_obj.port)
    global responder
    global server_should_shutdown
    server_should_shutdown = False
    responder = GenericResponder(proto, msg, datum)
    server = http_server.HTTPServer(server_addr, GenericHandler)
    print("Port: %s" % server.server_port)
    sys.stdout.flush()
    server.allow_reuse_address = True
    print("Starting server.", file=sys.stderr)
    server.serve_forever()


def send_message(uri, proto, msg, datum):
    url_obj = urlparse(uri)
    client = ipc.HTTPTransceiver(url_obj.hostname, url_obj.port)
    proto_json = open(proto, 'rb').read()
    requestor = ipc.Requestor(protocol.parse(proto_json), client)
    print(requestor.request(msg, datum))

##
# TODO: Replace this with fileinput()


def file_or_stdin(f):
    return sys.stdin if f == '-' else open(f, 'rb')


def main(args=sys.argv):
    if len(args) == 1:
        print("Usage: %s [dump|rpcreceive|rpcsend]" % args[0])
        return 1

    if args[1] == "dump":
        if len(args) != 3:
            print("Usage: %s dump input_file" % args[0])
            return 1
        for d in datafile.DataFileReader(file_or_stdin(args[2]), avro.io.DatumReader()):
            print(repr(d))
    elif args[1] == "rpcreceive":
        usage_str = "Usage: %s rpcreceive uri protocol_file " % args[0]
        usage_str += "message_name (-data d | -file f)"
        if len(args) not in [5, 7]:
            print(usage_str)
            return 1
        uri, proto, msg = args[2:5]
        datum = None
        if len(args) > 5:
            if args[5] == "-file":
                reader = open(args[6], 'rb')
                datum_reader = avro.io.DatumReader()
                dfr = datafile.DataFileReader(reader, datum_reader)
                datum = next(dfr)
            elif args[5] == "-data":
                print("JSON Decoder not yet implemented.")
                return 1
            else:
                print(usage_str)
                return 1
        run_server(uri, proto, msg, datum)
    elif args[1] == "rpcsend":
        usage_str = "Usage: %s rpcsend uri protocol_file " % args[0]
        usage_str += "message_name (-data d | -file f)"
        if len(args) not in [5, 7]:
            print(usage_str)
            return 1
        uri, proto, msg = args[2:5]
        datum = None
        if len(args) > 5:
            if args[5] == "-file":
                reader = open(args[6], 'rb')
                datum_reader = avro.io.DatumReader()
                dfr = datafile.DataFileReader(reader, datum_reader)
                datum = next(dfr)
            elif args[5] == "-data":
                print("JSON Decoder not yet implemented.")
                return 1
            else:
                print(usage_str)
                return 1
        send_message(uri, proto, msg, datum)
    return 0


if __name__ == "__main__":
    if os.path.dirname(avro.io.__file__) in sys.path:
        warnings.warn("Invoking avro/tool.py directly is likely to lead to a name collision "
                      "with the python io module. Try doing `python -m avro.tool` instead.")

    sys.exit(main(sys.argv))



#!/usr/bin/env python

##
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
# https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import, division, print_function

import collections
import distutils.spawn
import os
import platform
import shutil
import subprocess
import sys
import tempfile
import unittest

import avro
import avro.datafile
import avro.io
import avro.schema
import avro.tether.tether_task_runner

try:
    unicode
except NameError:
    unicode = str

_AVRO_DIR = os.path.abspath(os.path.dirname(avro.__file__))


def _version():
    with open(os.path.join(_AVRO_DIR, 'VERSION.txt')) as v:
        # Convert it back to the java version
        return v.read().strip().replace('+', '-')


_AVRO_VERSION = _version()
_JAR_PATH = os.path.join(os.path.dirname(os.path.dirname(_AVRO_DIR)),
                         "java", "tools", "target", "avro-tools-{}.jar".format(_AVRO_VERSION))

_LINES = (unicode("the quick brown fox jumps over the lazy dog"),
          unicode("the cow jumps over the moon"),
          unicode("the rain in spain falls mainly on the plains"))
_IN_SCHEMA = '"string"'

# The schema for the output of the mapper and reducer
_OUT_SCHEMA = """{
  "type": "record",
  "name": "Pair",
  "namespace": "org.apache.avro.mapred",
  "fields": [{"name": "key", "type": "string"},
             {"name": "value", "type": "long", "order": "ignore"}]
}"""

_PYTHON_PATH = os.pathsep.join([os.path.dirname(os.path.dirname(avro.__file__)),
                                os.path.dirname(__file__)])


def _has_java():
    """Detect if this system has a usable java installed.

    On most systems, this is just checking if `java` is in the PATH.

    But macos always has a /usr/bin/java, which does not mean java is installed.
    If you invoke java on macos and java is not installed, macos will spawn a popup
    telling you how to install java. This code does additional work around that
    to be completely automatic.
    """
    if platform.system() == "Darwin":
        try:
            output = subprocess.check_output("/usr/libexec/java_home", stderr=subprocess.STDOUT)
        except subprocess.CalledProcessError as e:
            output = e.output
        return (b"No Java runtime present" not in output)
    return bool(distutils.spawn.find_executable("java"))


@unittest.skipUnless(_has_java(), "No Java runtime present")
@unittest.skipUnless(os.path.exists(_JAR_PATH), "{} not found".format(_JAR_PATH))
class TestTetherWordCount(unittest.TestCase):
    """unittest for a python tethered map-reduce job."""

    _base_dir = None
    _script_path = None
    _input_path = None
    _output_path = None
    _output_schema_path = None

    def setUp(self):
        """Create temporary files for testing."""
        prefix, _ = os.path.splitext(os.path.basename(__file__))
        self._base_dir = tempfile.mkdtemp(prefix=prefix)

        # We create the input path...
        self._input_path = os.path.join(self._base_dir, "in")
        if not os.path.exists(self._input_path):
            os.makedirs(self._input_path)
        infile = os.path.join(self._input_path, "lines.avro")
        self._write_lines(_LINES, infile)
        self.assertTrue(os.path.exists(infile), "Missing the input file {}".format(infile))

        # ...and the output schema...
        self._output_schema_path = os.path.join(self._base_dir, "output.avsc")
        with open(self._output_schema_path, 'w') as output_schema_handle:
            output_schema_handle.write(_OUT_SCHEMA)
        self.assertTrue(os.path.exists(self._output_schema_path), "Missing the schema file")

        # ...but we just name the output path. The tether tool creates it.
        self._output_path = os.path.join(self._base_dir, "out")

    def tearDown(self):
        """Remove temporary files used in testing."""
        if os.path.exists(self._base_dir):
            shutil.rmtree(self._base_dir)
        if self._script_path is not None and os.path.exists(self._script_path):
            os.remove(self._script_path)

    def _write_lines(self, lines, fname):
        """
        Write the lines to an avro file named fname

        Parameters
        --------------------------------------------------------
        lines - list of strings to write
        fname - the name of the file to write to.
        """
        datum_writer = avro.io.DatumWriter(_IN_SCHEMA)
        writers_schema = avro.schema.parse(_IN_SCHEMA)
        with avro.datafile.DataFileWriter(open(fname, 'wb'), datum_writer, writers_schema) as writer:
            for datum in lines:
                writer.append(datum)

    def test_tether_word_count(self):
        """Check that a tethered map-reduce job produces the output expected locally."""
        # Run the job...
        args = ("java", "-jar", _JAR_PATH, "tether",
                "--protocol", "http",
                "--in", self._input_path,
                "--out", self._output_path,
                "--outschema", self._output_schema_path,
                "--program", sys.executable,
                "--exec_args", "-m avro.tether.tether_task_runner word_count_task.WordCountTask")
        print("Command:\n\t{0}".format(" ".join(args)))
        subprocess.check_call(args, env={"PYTHONPATH": _PYTHON_PATH, "PATH": os.environ["PATH"]})

        # ...and test the results.
        datum_reader = avro.io.DatumReader()
        outfile = os.path.join(self._output_path, "part-00000.avro")
        expected_counts = collections.Counter(' '.join(_LINES).split())
        with avro.datafile.DataFileReader(open(outfile, 'rb'), datum_reader) as reader:
            actual_counts = {r["key"]: r["value"] for r in reader}
        self.assertDictEqual(actual_counts, expected_counts)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python

##
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import, division, print_function

import logging
import sys
import threading
import traceback
import weakref

import avro.tether.tether_task
import avro.tether.util
from avro import ipc

try:
    import BaseHTTPServer as http_server  # type: ignore
except ImportError:
    import http.server as http_server  # type: ignore

__all__ = ["TaskRunner"]


class TaskRunnerResponder(ipc.Responder):
    """
    The responder for the tethered process
    """

    def __init__(self, runner):
        """
        Param
        ----------------------------------------------------------
        runner - Instance of TaskRunner
        """
        ipc.Responder.__init__(self, avro.tether.tether_task.inputProtocol)

        self.log = logging.getLogger("TaskRunnerResponder")

        # should we use weak references to avoid circular references?
        # We use weak references b\c self.runner owns this instance of TaskRunnerResponder
        if isinstance(runner, weakref.ProxyType):
            self.runner = runner
        else:
            self.runner = weakref.proxy(runner)

        self.task = weakref.proxy(runner.task)

    def invoke(self, message, request):
        try:
            if message.name == 'configure':
                self.log.info("TetherTaskRunner: Received configure")
                self.task.configure(request["taskType"], request["inSchema"], request["outSchema"])
            elif message.name == 'partitions':
                self.log.info("TetherTaskRunner: Received partitions")
                try:
                    self.task.partitions = request["partitions"]
                except Exception as e:
                    self.log.error("Exception occured while processing the partitions message: Message:\n" + traceback.format_exc())
                    raise
            elif message.name == 'input':
                self.log.info("TetherTaskRunner: Received input")
                self.task.input(request["data"], request["count"])
            elif message.name == 'abort':
                self.log.info("TetherTaskRunner: Received abort")
                self.runner.close()
            elif message.name == 'complete':
                self.log.info("TetherTaskRunner: Received complete")
                self.task.complete()
                self.task.close()
                self.runner.close()
            else:
                self.log.warning("TetherTaskRunner: Received unknown message {0}".format(message.name))

        except Exception as e:
            self.log.error("Error occured while processing message: {0}".format(message.name))
            emsg = traceback.format_exc()
            self.task.fail(emsg)

        return None


def HTTPHandlerGen(runner):
    """
    This is a class factory for the HTTPHandler. We need
    a factory because we need a reference to the runner

    Parameters
    -----------------------------------------------------------------
    runner - instance of the task runner
    """

    if not(isinstance(runner, weakref.ProxyType)):
        runnerref = weakref.proxy(runner)
    else:
        runnerref = runner

    class TaskRunnerHTTPHandler(http_server.BaseHTTPRequestHandler):
        """Create a handler for the parent.
        """

        runner = runnerref

        def __init__(self, *args, **param):
            """
            """
            http_server.BaseHTTPRequestHandler.__init__(self, *args, **param)

        def do_POST(self):
            self.responder = TaskRunnerResponder(self.runner)
            call_request_reader = ipc.FramedReader(self.rfile)
            call_request = call_request_reader.read_framed_message()
            resp_body = self.responder.respond(call_request)
            self.send_response(200)
            self.send_header('Content-Type', 'avro/binary')
            self.end_headers()
            resp_writer = ipc.FramedWriter(self.wfile)
            resp_writer.write_framed_message(resp_body)

    return TaskRunnerHTTPHandler


class TaskRunner(object):
    """This class ties together the server handling the requests from
    the parent process and the instance of TetherTask which actually
    implements the logic for the mapper and reducer phases
    """

    def __init__(self, task):
        """
        Construct the runner

        Parameters
        ---------------------------------------------------------------
        task - An instance of tether task
        """

        self.log = logging.getLogger("TaskRunner:")

        if not(isinstance(task, avro.tether.tether_task.TetherTask)):
            raise ValueError("task must be an instance of tether task")
        self.task = task

        self.server = None
        self.sthread = None

    def start(self, outputport=None, join=True):
        """
        Start the server

        Parameters
        -------------------------------------------------------------------
        outputport - (optional) The port on which the parent process is listening
                     for requests from the task.
                   - This will typically be supplied by an environment variable
                     we allow it to be supplied as an argument mainly for debugging
        join       - (optional) If set to fault then we don't issue a join to block
                     until the thread excecuting the server terminates.
                    This is mainly for debugging. By setting it to false,
                    we can resume execution in this thread so that we can do additional
                    testing
        """

        port = avro.tether.util.find_port()
        address = ("localhost", port)

        def thread_run(task_runner=None):
            task_runner.server = http_server.HTTPServer(address, HTTPHandlerGen(task_runner))
            task_runner.server.allow_reuse_address = True
            task_runner.server.serve_forever()

        # create a separate thread for the http server
        sthread = threading.Thread(target=thread_run, kwargs={"task_runner": self})
        sthread.start()

        self.sthread = sthread
        # This needs to run in a separat thread b\c serve_forever() blocks
        self.task.open(port, clientPort=outputport)

        # wait for the other thread to finish
        if (join):
            self.task.ready_for_shutdown.wait()
            self.server.shutdown()

            # should we do some kind of check to make sure it exits
            self.log.info("Shutdown the logger")
            # shutdown the logging
            logging.shutdown()

    def close(self):
        """
        Handler for the close message
        """

        self.task.close()


if __name__ == '__main__':
    # TODO::Make the logging level a parameter we can set
    # logging.basicConfig(level=logging.INFO,filename='/tmp/log',filemode='w')
    logging.basicConfig(level=logging.INFO)

    if (len(sys.argv) <= 1):
        print("Error: tether_task_runner.__main__: Usage: tether_task_runner task_package.task_module.TaskClass")
        raise ValueError("Usage: tether_task_runner task_package.task_module.TaskClass")

    fullcls = sys.argv[1]
    mod, cname = fullcls.rsplit(".", 1)

    logging.info("tether_task_runner.__main__: Task: {0}".format(fullcls))

    modobj = __import__(mod, fromlist=cname)

    taskcls = getattr(modobj, cname)
    task = taskcls()

    runner = TaskRunner(task=task)
    runner.start()



#!/usr/bin/env python

##
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
# https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Contains Constants for Python Avro"""

from __future__ import absolute_import, division, print_function

DATE = "date"
DECIMAL = "decimal"
TIMESTAMP_MICROS = "timestamp-micros"
TIMESTAMP_MILLIS = "timestamp-millis"
TIME_MICROS = "time-micros"
TIME_MILLIS = "time-millis"

SUPPORTED_LOGICAL_TYPE = [
    DATE,
    DECIMAL,
    TIMESTAMP_MICROS,
    TIMESTAMP_MILLIS,
    TIME_MICROS,
    TIME_MILLIS
]



#!/usr/bin/env python

##
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
# https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
There are currently no IPC tests within python, in part because there are no
servers yet available.
"""

from __future__ import absolute_import, division, print_function

import unittest

# This test does import this code, to make sure it at least passes
# compilation.
from avro import ipc


class TestIPC(unittest.TestCase):
    def test_placeholder(self):
        pass

    def test_server_with_path(self):
        client_with_custom_path = ipc.HTTPTransceiver('apache.org', 80, '/service/article')
        self.assertEqual('/service/article', client_with_custom_path.req_resource)

        client_with_default_path = ipc.HTTPTransceiver('apache.org', 80)
        self.assertEqual('/', client_with_default_path.req_resource)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python

##
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
# https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import, division, print_function

from datetime import datetime, timedelta, tzinfo


class UTCTzinfo(tzinfo):
    def utcoffset(self, dt):
        return timedelta(0)

    def tzname(self, dt):
        return "UTC"

    def dst(self, dt):
        return timedelta(0)


utc = UTCTzinfo()


# Test Time Zone with fixed offset and no DST
class TSTTzinfo(tzinfo):
    def utcoffset(self, dt):
        return timedelta(hours=10)

    def tzname(self, dt):
        return "TST"

    def dst(self, dt):
        return timedelta(0)


tst = TSTTzinfo()



#!/usr/bin/env python

##
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
# https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import, division, print_function

import io
import logging
import os
import subprocess
import sys
import time
import unittest

import avro.io
import avro.test.mock_tether_parent
import avro.test.word_count_task
import avro.tether.tether_task
import avro.tether.tether_task_runner
import avro.tether.util

try:
    unicode
except NameError:
    unicode = str


class TestTetherTaskRunner(unittest.TestCase):
    """unit test for a tethered task runner."""

    def test1(self):
        # set the logging level to debug so that debug messages are printed
        logging.basicConfig(level=logging.DEBUG)

        proc = None
        try:
            # launch the server in a separate process
            env = dict()
            env["PYTHONPATH"] = ':'.join(sys.path)
            parent_port = avro.tether.util.find_port()

            pyfile = avro.test.mock_tether_parent.__file__
            proc = subprocess.Popen([sys.executable, pyfile, "start_server", "{0}".format(parent_port)])
            input_port = avro.tether.util.find_port()

            print("Mock server started process pid={0}".format(proc.pid))
            # Possible race condition? open tries to connect to the subprocess before the subprocess is fully started
            # so we give the subprocess time to start up
            time.sleep(1)

            runner = avro.tether.tether_task_runner.TaskRunner(avro.test.word_count_task.WordCountTask())

            runner.start(outputport=parent_port, join=False)

            # Test sending various messages to the server and ensuring they are processed correctly
            requestor = avro.tether.tether_task.HTTPRequestor(
                "localhost", runner.server.server_address[1], avro.tether.tether_task.inputProtocol)

            # TODO: We should validate that open worked by grabbing the STDOUT of the subproces
            # and ensuring that it outputted the correct message.

            # Test the mapper
            requestor.request("configure", {
                "taskType": avro.tether.tether_task.TaskType.MAP,
                "inSchema": unicode(str(runner.task.inschema)),
                "outSchema": unicode(str(runner.task.midschema))
            })

            # Serialize some data so we can send it to the input function
            datum = unicode("This is a line of text")
            writer = io.BytesIO()
            encoder = avro.io.BinaryEncoder(writer)
            datum_writer = avro.io.DatumWriter(runner.task.inschema)
            datum_writer.write(datum, encoder)

            writer.seek(0)
            data = writer.read()

            # Call input to simulate calling map
            requestor.request("input", {"data": data, "count": 1})

            # Test the reducer
            requestor.request("configure", {
                "taskType": avro.tether.tether_task.TaskType.REDUCE,
                "inSchema": unicode(str(runner.task.midschema)),
                "outSchema": unicode(str(runner.task.outschema))}
            )

            # Serialize some data so we can send it to the input function
            datum = {"key": unicode("word"), "value": 2}
            writer = io.BytesIO()
            encoder = avro.io.BinaryEncoder(writer)
            datum_writer = avro.io.DatumWriter(runner.task.midschema)
            datum_writer.write(datum, encoder)

            writer.seek(0)
            data = writer.read()

            # Call input to simulate calling reduce
            requestor.request("input", {"data": data, "count": 1})

            requestor.request("complete", {})

            runner.task.ready_for_shutdown.wait()
            runner.server.shutdown()
            # time.sleep(2)
            # runner.server.shutdown()

            sthread = runner.sthread

            # Possible race condition?
            time.sleep(1)

            # make sure the other thread terminated
            self.assertFalse(sthread.isAlive())

            # shutdown the logging
            logging.shutdown()

        except Exception as e:
            raise
        finally:
            # close the process
            if not(proc is None):
                proc.kill()

    def test2(self):
        """
        In this test we want to make sure that when we run "tether_task_runner.py"
        as our main script everything works as expected. We do this by using subprocess to run it
        in a separate thread.
        """
        proc = None

        runnerproc = None
        try:
            # launch the server in a separate process
            env = dict()
            env["PYTHONPATH"] = ':'.join(sys.path)
            parent_port = avro.tether.util.find_port()

            pyfile = avro.test.mock_tether_parent.__file__
            proc = subprocess.Popen([sys.executable, pyfile, "start_server", "{0}".format(parent_port)])

            # Possible race condition? when we start tether_task_runner it will call
            # open tries to connect to the subprocess before the subprocess is fully started
            # so we give the subprocess time to start up
            time.sleep(1)

            # start the tether_task_runner in a separate process
            env = {"AVRO_TETHER_OUTPUT_PORT": "{0}".format(parent_port)}
            env["PYTHONPATH"] = ':'.join(sys.path)

            runnerproc = subprocess.Popen([sys.executable, avro.tether.tether_task_runner.__file__,
                                          "avro.test.word_count_task.WordCountTask"], env=env)

            # possible race condition wait for the process to start
            time.sleep(1)

            print("Mock server started process pid={0}".format(proc.pid))
            # Possible race condition? open tries to connect to the subprocess before the subprocess is fully started
            # so we give the subprocess time to start up
            time.sleep(1)

        except Exception as e:
            raise
        finally:
            # close the process
            if not(runnerproc is None):
                runnerproc.kill()

            if not(proc is None):
                proc.kill()


if __name__ == ("__main__"):
    unittest.main()



#!/usr/bin/env python

##
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
# https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import, division, print_function

import os
import unittest

import avro
from avro import datafile, io

_INTEROP_DATA_DIR = os.path.join(os.path.dirname(avro.__file__), 'test', 'interop', 'data')


@unittest.skipUnless(os.path.exists(_INTEROP_DATA_DIR),
                     "{} does not exist".format(_INTEROP_DATA_DIR))
class TestDataFileInterop(unittest.TestCase):
    def test_interop(self):
        """Test Interop"""
        for f in os.listdir(_INTEROP_DATA_DIR):
            filename = os.path.join(_INTEROP_DATA_DIR, f)
            assert os.stat(filename).st_size > 0
            base_ext = os.path.splitext(os.path.basename(f))[0].split('_', 1)
            if len(base_ext) < 2 or base_ext[1] in datafile.VALID_CODECS:
                print('READING %s' % f)
                print()

                # read data in binary from file
                datum_reader = io.DatumReader()
                with open(filename, 'rb') as reader:
                    dfr = datafile.DataFileReader(reader, datum_reader)
                    i = 0
                    for i, datum in enumerate(dfr, 1):
                        assert datum is not None
                    assert i > 0
            else:
                print('SKIPPING %s due to an unsupported codec' % f)
                print()


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python

##
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
# https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import, division, print_function

import datetime
import io
import unittest
from binascii import hexlify
from decimal import Decimal

import avro.io
from avro import schema, timezones

try:
    unicode
except NameError:
    unicode = str

SCHEMAS_TO_VALIDATE = (
    ('"null"', None),
    ('"boolean"', True),
    ('"string"', unicode('adsfasdf09809dsf-=adsf')),
    ('"bytes"', b'12345abcd'),
    ('"int"', 1234),
    ('"long"', 1234),
    ('"float"', 1234.0),
    ('"double"', 1234.0),
    ('{"type": "fixed", "name": "Test", "size": 1}', b'B'),
    ('{"type": "fixed", "logicalType": "decimal", "name": "Test", "size": 8, "precision": 5, "scale": 4}',
     Decimal('3.1415')),
    ('{"type": "fixed", "logicalType": "decimal", "name": "Test", "size": 8, "precision": 5, "scale": 4}',
     Decimal('-3.1415')),
    ('{"type": "bytes", "logicalType": "decimal", "precision": 5, "scale": 4}', Decimal('3.1415')),
    ('{"type": "bytes", "logicalType": "decimal", "precision": 5, "scale": 4}', Decimal('-3.1415')),
    ('{"type": "enum", "name": "Test", "symbols": ["A", "B"]}', 'B'),
    ('{"type": "array", "items": "long"}', [1, 3, 2]),
    ('{"type": "map", "values": "long"}', {unicode('a'): 1,
                                           unicode('b'): 3,
                                           unicode('c'): 2}),
    ('["string", "null", "long"]', None),
    ('{"type": "int", "logicalType": "date"}', datetime.date(2000, 1, 1)),
    ('{"type": "int", "logicalType": "time-millis"}', datetime.time(23, 59, 59, 999000)),
    ('{"type": "int", "logicalType": "time-millis"}', datetime.time(0, 0, 0, 000000)),
    ('{"type": "long", "logicalType": "time-micros"}', datetime.time(23, 59, 59, 999999)),
    ('{"type": "long", "logicalType": "time-micros"}', datetime.time(0, 0, 0, 000000)),
    (
        '{"type": "long", "logicalType": "timestamp-millis"}',
        datetime.datetime(1000, 1, 1, 0, 0, 0, 000000, tzinfo=timezones.utc)
    ),
    (
        '{"type": "long", "logicalType": "timestamp-millis"}',
        datetime.datetime(9999, 12, 31, 23, 59, 59, 999000, tzinfo=timezones.utc)
    ),
    (
        '{"type": "long", "logicalType": "timestamp-millis"}',
        datetime.datetime(2000, 1, 18, 2, 2, 1, 100000, tzinfo=timezones.tst)
    ),
    (
        '{"type": "long", "logicalType": "timestamp-micros"}',
        datetime.datetime(1000, 1, 1, 0, 0, 0, 000000, tzinfo=timezones.utc)
    ),
    (
        '{"type": "long", "logicalType": "timestamp-micros"}',
        datetime.datetime(9999, 12, 31, 23, 59, 59, 999999, tzinfo=timezones.utc)
    ),
    (
        '{"type": "long", "logicalType": "timestamp-micros"}',
        datetime.datetime(2000, 1, 18, 2, 2, 1, 123499, tzinfo=timezones.tst)
    ),
    ('{"type": "string", "logicalType": "uuid"}', u'12345abcd'),
    ('{"type": "string", "logicalType": "unknown-logical-type"}', u'12345abcd'),
    ('{"type": "string", "logicalType": "timestamp-millis"}', u'12345abcd'),
    ("""\
   {"type": "record",
    "name": "Test",
    "fields": [{"name": "f", "type": "long"}]}
   """, {'f': 5}),
    ("""\
   {"type": "record",
    "name": "Lisp",
    "fields": [{"name": "value",
                "type": ["null", "string",
                         {"type": "record",
                          "name": "Cons",
                          "fields": [{"name": "car", "type": "Lisp"},
                                     {"name": "cdr", "type": "Lisp"}]}]}]}
   """, {'value': {'car': {'value': unicode('head')}, 'cdr': {'value': None}}}),
)

BINARY_ENCODINGS = (
    (0, b'00'),
    (-1, b'01'),
    (1, b'02'),
    (-2, b'03'),
    (2, b'04'),
    (-64, b'7f'),
    (64, b'80 01'),
    (8192, b'80 80 01'),
    (-8193, b'81 80 01'),
)

DEFAULT_VALUE_EXAMPLES = (
    ('"null"', 'null', None),
    ('"boolean"', 'true', True),
    ('"string"', '"foo"', u'foo'),
    ('"bytes"', '"\u00FF\u00FF"', u'\xff\xff'),
    ('"int"', '5', 5),
    ('"long"', '5', 5),
    ('"float"', '1.1', 1.1),
    ('"double"', '1.1', 1.1),
    ('{"type": "fixed", "name": "F", "size": 2}', '"\u00FF\u00FF"', u'\xff\xff'),
    ('{"type": "enum", "name": "F", "symbols": ["FOO", "BAR"]}', '"FOO"', 'FOO'),
    ('{"type": "array", "items": "int"}', '[1, 2, 3]', [1, 2, 3]),
    ('{"type": "map", "values": "int"}', '{"a": 1, "b": 2}', {unicode('a'): 1,
                                                              unicode('b'): 2}),
    ('["int", "null"]', '5', 5),
    ('{"type": "record", "name": "F", "fields": [{"name": "A", "type": "int"}]}',
     '{"A": 5}', {'A': 5}),
)

LONG_RECORD_SCHEMA = schema.parse("""\
  {"type": "record",
   "name": "Test",
   "fields": [{"name": "A", "type": "int"},
              {"name": "B", "type": "int"},
              {"name": "C", "type": "int"},
              {"name": "D", "type": "int"},
              {"name": "E", "type": "int"},
              {"name": "F", "type": "int"},
              {"name": "G", "type": "int"}]}""")

LONG_RECORD_DATUM = {'A': 1, 'B': 2, 'C': 3, 'D': 4, 'E': 5, 'F': 6, 'G': 7}


def avro_hexlify(reader):
    """Return the hex value, as a string, of a binary-encoded int or long."""
    b = []
    current_byte = reader.read(1)
    b.append(hexlify(current_byte))
    while (ord(current_byte) & 0x80) != 0:
        current_byte = reader.read(1)
        b.append(hexlify(current_byte))
    return b' '.join(b)


def print_test_name(test_name):
    print('')
    print(test_name)
    print('=' * len(test_name))
    print('')


def write_datum(datum, writers_schema):
    writer = io.BytesIO()
    encoder = avro.io.BinaryEncoder(writer)
    datum_writer = avro.io.DatumWriter(writers_schema)
    datum_writer.write(datum, encoder)
    return writer, encoder, datum_writer


def read_datum(buffer, writers_schema, readers_schema=None):
    reader = io.BytesIO(buffer.getvalue())
    decoder = avro.io.BinaryDecoder(reader)
    datum_reader = avro.io.DatumReader(writers_schema, readers_schema)
    return datum_reader.read(decoder)


def check_binary_encoding(number_type):
    print_test_name('TEST BINARY %s ENCODING' % number_type.upper())
    correct = 0
    for datum, hex_encoding in BINARY_ENCODINGS:
        print('Datum: %d' % datum)
        print('Correct Encoding: %s' % hex_encoding)

        writers_schema = schema.parse('"%s"' % number_type.lower())
        writer, encoder, datum_writer = write_datum(datum, writers_schema)
        writer.seek(0)
        hex_val = avro_hexlify(writer)

        print('Read Encoding: %s' % hex_val)
        if hex_encoding == hex_val:
            correct += 1
        print('')
    return correct


def check_skip_number(number_type):
    print_test_name('TEST SKIP %s' % number_type.upper())
    correct = 0
    for value_to_skip, hex_encoding in BINARY_ENCODINGS:
        VALUE_TO_READ = 6253
        print('Value to Skip: %d' % value_to_skip)

        # write the value to skip and a known value
        writers_schema = schema.parse('"%s"' % number_type.lower())
        writer, encoder, datum_writer = write_datum(value_to_skip, writers_schema)
        datum_writer.write(VALUE_TO_READ, encoder)

        # skip the value
        reader = io.BytesIO(writer.getvalue())
        decoder = avro.io.BinaryDecoder(reader)
        decoder.skip_long()

        # read data from string buffer
        datum_reader = avro.io.DatumReader(writers_schema)
        read_value = datum_reader.read(decoder)

        print('Read Value: %d' % read_value)
        if read_value == VALUE_TO_READ:
            correct += 1
        print('')
    return correct


class TestIO(unittest.TestCase):
    #
    # BASIC FUNCTIONALITY
    #

    def test_validate(self):
        print_test_name('TEST VALIDATE')
        passed = 0
        for example_schema, datum in SCHEMAS_TO_VALIDATE:
            print('Schema: %s' % example_schema)
            print('Datum: %s' % datum)
            validated = avro.io.validate(schema.parse(example_schema), datum)
            print('Valid: %s' % validated)
            if validated:
                passed += 1
        self.assertEquals(passed, len(SCHEMAS_TO_VALIDATE))

    def test_round_trip(self):
        print_test_name('TEST ROUND TRIP')
        correct = 0
        for example_schema, datum in SCHEMAS_TO_VALIDATE:
            print('Schema: %s' % example_schema)
            print('Datum: %s' % datum)

            writers_schema = schema.parse(example_schema)
            writer, encoder, datum_writer = write_datum(datum, writers_schema)
            round_trip_datum = read_datum(writer, writers_schema)

            print('Round Trip Datum: %s' % round_trip_datum)
            if isinstance(round_trip_datum, Decimal):
                round_trip_datum = round_trip_datum.to_eng_string()
                datum = str(datum)
            elif isinstance(round_trip_datum, datetime.datetime):
                datum = datum.astimezone(tz=timezones.utc)
            if datum == round_trip_datum:
                correct += 1
        self.assertEquals(correct, len(SCHEMAS_TO_VALIDATE))

    #
    # BINARY ENCODING OF INT AND LONG
    #

    def test_binary_int_encoding(self):
        correct = check_binary_encoding('int')
        self.assertEquals(correct, len(BINARY_ENCODINGS))

    def test_binary_long_encoding(self):
        correct = check_binary_encoding('long')
        self.assertEquals(correct, len(BINARY_ENCODINGS))

    def test_skip_int(self):
        correct = check_skip_number('int')
        self.assertEquals(correct, len(BINARY_ENCODINGS))

    def test_skip_long(self):
        correct = check_skip_number('long')
        self.assertEquals(correct, len(BINARY_ENCODINGS))

    #
    # SCHEMA RESOLUTION
    #

    def test_schema_promotion(self):
        print_test_name('TEST SCHEMA PROMOTION')
        # note that checking writers_schema.type in read_data
        # allows us to handle promotion correctly
        promotable_schemas = ['"int"', '"long"', '"float"', '"double"']
        incorrect = 0
        for i, ws in enumerate(promotable_schemas):
            writers_schema = schema.parse(ws)
            datum_to_write = 219
            for rs in promotable_schemas[i + 1:]:
                readers_schema = schema.parse(rs)
                writer, enc, dw = write_datum(datum_to_write, writers_schema)
                datum_read = read_datum(writer, writers_schema, readers_schema)
                print('Writer: %s Reader: %s' % (writers_schema, readers_schema))
                print('Datum Read: %s' % datum_read)
                if datum_read != datum_to_write:
                    incorrect += 1
        self.assertEquals(incorrect, 0)

    def test_unknown_symbol(self):
        print_test_name('TEST UNKNOWN SYMBOL')
        writers_schema = schema.parse("""\
      {"type": "enum", "name": "Test",
       "symbols": ["FOO", "BAR"]}""")
        datum_to_write = 'FOO'

        readers_schema = schema.parse("""\
      {"type": "enum", "name": "Test",
       "symbols": ["BAR", "BAZ"]}""")

        writer, encoder, datum_writer = write_datum(datum_to_write, writers_schema)
        reader = io.BytesIO(writer.getvalue())
        decoder = avro.io.BinaryDecoder(reader)
        datum_reader = avro.io.DatumReader(writers_schema, readers_schema)
        self.assertRaises(avro.io.SchemaResolutionException, datum_reader.read, decoder)

    def test_default_value(self):
        print_test_name('TEST DEFAULT VALUE')
        writers_schema = LONG_RECORD_SCHEMA
        datum_to_write = LONG_RECORD_DATUM

        correct = 0
        for field_type, default_json, default_datum in DEFAULT_VALUE_EXAMPLES:
            readers_schema = schema.parse("""\
        {"type": "record", "name": "Test",
         "fields": [{"name": "H", "type": %s, "default": %s}]}
        """ % (field_type, default_json))
            datum_to_read = {'H': default_datum}

            writer, encoder, datum_writer = write_datum(datum_to_write, writers_schema)
            datum_read = read_datum(writer, writers_schema, readers_schema)
            print('Datum Read: %s' % datum_read)
            if datum_to_read == datum_read:
                correct += 1
        self.assertEquals(correct, len(DEFAULT_VALUE_EXAMPLES))

    def test_no_default_value(self):
        print_test_name('TEST NO DEFAULT VALUE')
        writers_schema = LONG_RECORD_SCHEMA
        datum_to_write = LONG_RECORD_DATUM

        readers_schema = schema.parse("""\
      {"type": "record", "name": "Test",
       "fields": [{"name": "H", "type": "int"}]}""")

        writer, encoder, datum_writer = write_datum(datum_to_write, writers_schema)
        reader = io.BytesIO(writer.getvalue())
        decoder = avro.io.BinaryDecoder(reader)
        datum_reader = avro.io.DatumReader(writers_schema, readers_schema)
        self.assertRaises(avro.io.SchemaResolutionException, datum_reader.read, decoder)

    def test_projection(self):
        print_test_name('TEST PROJECTION')
        writers_schema = LONG_RECORD_SCHEMA
        datum_to_write = LONG_RECORD_DATUM

        readers_schema = schema.parse("""\
      {"type": "record", "name": "Test",
       "fields": [{"name": "E", "type": "int"},
                  {"name": "F", "type": "int"}]}""")
        datum_to_read = {'E': 5, 'F': 6}

        writer, encoder, datum_writer = write_datum(datum_to_write, writers_schema)
        datum_read = read_datum(writer, writers_schema, readers_schema)
        print('Datum Read: %s' % datum_read)
        self.assertEquals(datum_to_read, datum_read)

    def test_field_order(self):
        print_test_name('TEST FIELD ORDER')
        writers_schema = LONG_RECORD_SCHEMA
        datum_to_write = LONG_RECORD_DATUM

        readers_schema = schema.parse("""\
      {"type": "record", "name": "Test",
       "fields": [{"name": "F", "type": "int"},
                  {"name": "E", "type": "int"}]}""")
        datum_to_read = {'E': 5, 'F': 6}

        writer, encoder, datum_writer = write_datum(datum_to_write, writers_schema)
        datum_read = read_datum(writer, writers_schema, readers_schema)
        print('Datum Read: %s' % datum_read)
        self.assertEquals(datum_to_read, datum_read)

    def test_type_exception(self):
        print_test_name('TEST TYPE EXCEPTION')
        writers_schema = schema.parse("""\
      {"type": "record", "name": "Test",
       "fields": [{"name": "F", "type": "int"},
                  {"name": "E", "type": "int"}]}""")
        datum_to_write = {'E': 5, 'F': 'Bad'}
        self.assertRaises(avro.io.AvroTypeException, write_datum, datum_to_write, writers_schema)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python

##
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
# https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Test the protocol parsing logic."""

from __future__ import absolute_import, division, print_function

import json
import unittest

import avro.protocol
import avro.schema

try:
    unicode
except NameError:
    unicode = str

try:
    basestring  # type: ignore
except NameError:
    basestring = (bytes, unicode)


class TestProtocol(object):
    """A proxy for a protocol string that provides useful test metadata."""

    def __init__(self, data, name='', comment=''):
        if not isinstance(data, basestring):
            data = json.dumps(data)
        self.data = data
        self.name = name or data
        self.comment = comment

    def parse(self):
        return avro.protocol.parse(str(self))

    def __str__(self):
        return str(self.data)


class ValidTestProtocol(TestProtocol):
    """A proxy for a valid protocol string that provides useful test metadata."""
    valid = True


class InvalidTestProtocol(TestProtocol):
    """A proxy for an invalid protocol string that provides useful test metadata."""
    valid = False


HELLO_WORLD = ValidTestProtocol({
    "namespace": "com.acme",
    "protocol": "HelloWorld",
    "types": [
        {"name": "Greeting", "type": "record", "fields": [
            {"name": "message", "type": "string"}]},
        {"name": "Curse", "type": "error", "fields": [
            {"name": "message", "type": "string"}]}
    ],
    "messages": {
        "hello": {
            "request": [{"name": "greeting", "type": "Greeting"}],
            "response": "Greeting",
            "errors": ["Curse"]
        }
    }
})
EXAMPLES = [HELLO_WORLD, ValidTestProtocol({
    "namespace": "org.apache.avro.test",
    "protocol": "Simple",
    "types": [
        {"name": "Kind", "type": "enum", "symbols": ["FOO", "BAR", "BAZ"]},
        {"name": "MD5", "type": "fixed", "size": 16},
        {"name": "TestRecord", "type": "record", "fields": [
            {"name": "name", "type": "string", "order": "ignore"},
            {"name": "kind", "type": "Kind", "order": "descending"},
            {"name": "hash", "type": "MD5"}
        ]},
        {"name": "TestError", "type": "error", "fields": [{"name": "message", "type": "string"}]}
    ],
    "messages": {
        "hello": {
            "request": [{"name": "greeting", "type": "string"}],
            "response": "string"
        }, "echo": {
            "request": [{"name": "record", "type": "TestRecord"}],
            "response": "TestRecord"
        }, "add": {
            "request": [{"name": "arg1", "type": "int"}, {"name": "arg2", "type": "int"}],
            "response": "int"
        }, "echoBytes": {
            "request": [{"name": "data", "type": "bytes"}],
            "response": "bytes"
        }, "error": {
            "request": [],
            "response": "null",
            "errors": ["TestError"]
        }
    }
}), ValidTestProtocol({
    "namespace": "org.apache.avro.test.namespace",
    "protocol": "TestNamespace",
    "types": [
        {"name": "org.apache.avro.test.util.MD5", "type": "fixed", "size": 16},
        {"name": "TestRecord", "type": "record", "fields": [
            {"name": "hash", "type": "org.apache.avro.test.util.MD5"}
        ]},
        {"name": "TestError", "namespace": "org.apache.avro.test.errors", "type": "error",
         "fields": [{"name": "message", "type": "string"}]}
    ],
    "messages": {
        "echo": {
            "request": [{"name": "record", "type": "TestRecord"}],
            "response": "TestRecord"
        }, "error": {
            "request": [],
            "response": "null",
            "errors": ["org.apache.avro.test.errors.TestError"]
        }
    }
}), ValidTestProtocol({
    "namespace": "org.apache.avro.test.namespace",
    "protocol": "TestImplicitNamespace",
    "types": [
                {"name": "org.apache.avro.test.util.MD5", "type": "fixed", "size": 16},
                {"name": "ReferencedRecord", "type": "record",
                 "fields": [{"name": "foo", "type": "string"}]},
                {"name": "TestRecord", "type": "record",
                 "fields": [{"name": "hash", "type": "org.apache.avro.test.util.MD5"},
                            {"name": "unqualified", "type": "ReferencedRecord"}]
                 },
                {"name": "TestError", "type": "error", "fields": [{"name": "message", "type": "string"}]}
    ],
    "messages": {
        "echo": {
            "request": [{"name": "qualified", "type": "org.apache.avro.test.namespace.TestRecord"}],
            "response": "TestRecord"
        }, "error": {
            "request": [],
            "response": "null",
            "errors": ["org.apache.avro.test.namespace.TestError"]
        }
    }
}), ValidTestProtocol({
    "namespace": "org.apache.avro.test.namespace",
    "protocol": "TestNamespaceTwo",
    "types": [
        {"name": "org.apache.avro.test.util.MD5", "type": "fixed", "size": 16},
        {"name": "ReferencedRecord", "type": "record",
         "namespace": "org.apache.avro.other.namespace",
         "fields": [{"name": "foo", "type": "string"}]},
        {"name": "TestRecord", "type": "record",
         "fields": [{"name": "hash", "type": "org.apache.avro.test.util.MD5"},
                    {"name": "qualified",
                     "type": "org.apache.avro.other.namespace.ReferencedRecord"}]
         },
        {"name": "TestError",
         "type": "error", "fields": [{"name": "message", "type": "string"}]}],
    "messages": {
        "echo": {
            "request": [{"name": "qualified", "type": "org.apache.avro.test.namespace.TestRecord"}],
            "response": "TestRecord"
        }, "error": {
            "request": [],
            "response": "null",
                        "errors": ["org.apache.avro.test.namespace.TestError"]
        }
    }
}), ValidTestProtocol({
    "namespace": "org.apache.avro.test.namespace",
    "protocol": "TestValidRepeatedName",
    "types": [
        {"name": "org.apache.avro.test.util.MD5", "type": "fixed", "size": 16},
        {"name": "ReferencedRecord", "type": "record",
         "namespace": "org.apache.avro.other.namespace",
         "fields": [{"name": "foo", "type": "string"}]},
        {"name": "ReferencedRecord", "type": "record",
         "fields": [{"name": "bar", "type": "double"}]},
        {"name": "TestError",
         "type": "error", "fields": [{"name": "message", "type": "string"}]}],
    "messages": {
        "echo": {
            "request": [{"name": "qualified", "type": "ReferencedRecord"}],
            "response": "org.apache.avro.other.namespace.ReferencedRecord"},
        "error": {
            "request": [],
            "response": "null",
            "errors": ["org.apache.avro.test.namespace.TestError"]}
    }
}), InvalidTestProtocol({
    "namespace": "org.apache.avro.test.namespace",
    "protocol": "TestInvalidRepeatedName",
    "types": [
        {"name": "org.apache.avro.test.util.MD5", "type": "fixed", "size": 16},
        {"name": "ReferencedRecord", "type": "record",
         "fields": [{"name": "foo", "type": "string"}]},
        {"name": "ReferencedRecord", "type": "record",
         "fields": [{"name": "bar", "type": "double"}]},
        {"name": "TestError",
         "type": "error", "fields": [{"name": "message", "type": "string"}]}],
    "messages": {
        "echo": {
            "request": [{"name": "qualified", "type": "ReferencedRecord"}],
            "response": "org.apache.avro.other.namespace.ReferencedRecord"
        }, "error": {
            "request": [],
            "response": "null",
            "errors": ["org.apache.avro.test.namespace.TestError"]
        }
    }
}),
    ValidTestProtocol({
        "namespace": "org.apache.avro.test",
        "protocol": "BulkData",
        "types": [],
        "messages": {
            "read": {
                "request": [],
                "response": "bytes"
            }, "write": {
                "request": [{"name": "data", "type": "bytes"}],
                "response": "null"
            }
        }
    }), ValidTestProtocol({
        "protocol": "API",
        "namespace": "xyz.api",
        "types": [{
            "type": "enum",
            "name": "Symbology",
            "namespace": "xyz.api.product",
            "symbols": ["OPRA", "CUSIP", "ISIN", "SEDOL"]
        }, {
            "type": "record",
            "name": "Symbol",
            "namespace": "xyz.api.product",
            "fields": [{"name": "symbology", "type": "xyz.api.product.Symbology"},
                       {"name": "symbol", "type": "string"}]
        }, {
            "type": "record",
            "name": "MultiSymbol",
            "namespace": "xyz.api.product",
            "fields": [{"name": "symbols",
                        "type": {"type": "map", "values": "xyz.api.product.Symbol"}}]
        }],
        "messages": {}
    }),
]

VALID_EXAMPLES = [e for e in EXAMPLES if getattr(e, "valid", False)]


class TestMisc(unittest.TestCase):
    def test_inner_namespace_set(self):
        print('')
        print('TEST INNER NAMESPACE')
        print('===================')
        print('')
        proto = HELLO_WORLD.parse()
        self.assertEqual(proto.namespace, "com.acme")
        self.assertEqual(proto.fullname, "com.acme.HelloWorld")
        greeting_type = proto.types_dict['Greeting']
        self.assertEqual(greeting_type.namespace, 'com.acme')

    def test_inner_namespace_not_rendered(self):
        proto = HELLO_WORLD.parse()
        self.assertEqual('com.acme.Greeting', proto.types[0].fullname)
        self.assertEqual('Greeting', proto.types[0].name)
        # but there shouldn't be 'namespace' rendered to json on the inner type
        self.assertFalse('namespace' in proto.to_json()['types'][0])


class ProtocolParseTestCase(unittest.TestCase):
    """Enable generating parse test cases over all the valid and invalid example protocols."""

    def __init__(self, test_proto):
        """Ignore the normal signature for unittest.TestCase because we are generating
        many test cases from this one class. This is safe as long as the autoloader
        ignores this class. The autoloader will ignore this class as long as it has
        no methods starting with `test_`.
        """
        super(ProtocolParseTestCase, self).__init__(
            'parse_valid' if test_proto.valid else 'parse_invalid')
        self.test_proto = test_proto

    def parse_valid(self):
        """Parsing a valid protocol should not error."""
        try:
            self.test_proto.parse()
        except avro.protocol.ProtocolParseException:
            self.fail("Valid protocol failed to parse: {!s}".format(self.test_proto))

    def parse_invalid(self):
        """Parsing an invalid schema should error."""
        try:
            self.test_proto.parse()
        except (avro.protocol.ProtocolParseException, avro.schema.SchemaParseException):
            pass
        else:
            self.fail("Invalid protocol should not have parsed: {!s}".format(self.test_proto))


class ErrorSchemaTestCase(unittest.TestCase):
    """Enable generating error schema test cases across all the valid test protocols."""

    def __init__(self, test_proto):
        """Ignore the normal signature for unittest.TestCase because we are generating
        many test cases from this one class. This is safe as long as the autoloader
        ignores this class. The autoloader will ignore this class as long as it has
        no methods starting with `test_`.
        """
        super(ErrorSchemaTestCase, self).__init__('check_error_schema_exists')
        self.test_proto = test_proto

    def check_error_schema_exists(self):
        """Protocol messages should always have at least a string error schema."""
        p = self.test_proto.parse()
        for k, m in p.messages.items():
            self.assertIsNotNone(m.errors, "Message {} did not have the expected implicit "
                                           "string error schema.".format(k))


class RoundTripParseTestCase(unittest.TestCase):
    """Enable generating round-trip parse test cases over all the valid test protocols."""

    def __init__(self, test_proto):
        """Ignore the normal signature for unittest.TestCase because we are generating
        many test cases from this one class. This is safe as long as the autoloader
        ignores this class. The autoloader will ignore this class as long as it has
        no methods starting with `test_`.
        """
        super(RoundTripParseTestCase, self).__init__('parse_round_trip')
        self.test_proto = test_proto

    def parse_round_trip(self):
        """The string of a Schema should be parseable to the same Schema."""
        parsed = self.test_proto.parse()
        round_trip = avro.protocol.parse(str(parsed))
        self.assertEqual(parsed, round_trip)


def load_tests(loader, default_tests, pattern):
    """Generate test cases across many test schema."""
    suite = unittest.TestSuite()
    suite.addTests(loader.loadTestsFromTestCase(TestMisc))
    suite.addTests(ProtocolParseTestCase(ex) for ex in EXAMPLES)
    suite.addTests(RoundTripParseTestCase(ex) for ex in VALID_EXAMPLES)
    return suite


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python

##
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
# https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Contains the Schema classes.

A schema may be one of:
  A record, mapping field names to field value data;
  An error, equivalent to a record;
  An enum, containing one of a small set of symbols;
  An array of values, all of the same schema;
  A map containing string/value pairs, each of a declared schema;
  A union of other schemas;
  A fixed sized binary object;
  A unicode string;
  A sequence of bytes;
  A 32-bit signed int;
  A 64-bit signed long;
  A 32-bit floating-point float;
  A 64-bit floating-point double;
  A boolean; or
  Null.
"""

from __future__ import absolute_import, division, print_function

import json
import math
import re
import sys
import warnings

from avro import constants

try:
    unicode
except NameError:
    unicode = str

try:
    basestring  # type: ignore
except NameError:
    basestring = (bytes, unicode)

#
# Constants
#

# The name portion of a fullname, record field names, and enum symbols must:
# start with [A-Za-z_]
# subsequently contain only [A-Za-z0-9_]
_BASE_NAME_PATTERN = re.compile(r'(?:^|\.)[A-Za-z_][A-Za-z0-9_]*$')

PRIMITIVE_TYPES = (
    'null',
    'boolean',
    'string',
    'bytes',
    'int',
    'long',
    'float',
    'double',
)

NAMED_TYPES = (
    'fixed',
    'enum',
    'record',
    'error',
)

VALID_TYPES = PRIMITIVE_TYPES + NAMED_TYPES + (
    'array',
    'map',
    'union',
    'request',
    'error_union'
)

SCHEMA_RESERVED_PROPS = (
    'type',
    'name',
    'namespace',
    'fields',     # Record
    'items',      # Array
    'size',       # Fixed
    'symbols',    # Enum
    'values',     # Map
    'doc',
)

FIELD_RESERVED_PROPS = (
    'default',
    'name',
    'doc',
    'order',
    'type',
)

VALID_FIELD_SORT_ORDERS = (
    'ascending',
    'descending',
    'ignore',
)

#
# Exceptions
#


class AvroException(Exception):
    pass


class SchemaParseException(AvroException):
    pass


class InvalidName(SchemaParseException):
    """User attempted to parse a schema with an invalid name."""


class AvroWarning(UserWarning):
    """Base class for warnings."""


class IgnoredLogicalType(AvroWarning):
    """Warnings for unknown or invalid logical types."""


def validate_basename(basename):
    """Raise InvalidName if the given basename is not a valid name."""
    if not _BASE_NAME_PATTERN.search(basename):
        raise InvalidName("{!s} is not a valid Avro name because it "
                          "does not match the pattern {!s}".format(
                              basename, _BASE_NAME_PATTERN.pattern))

#
# Base Classes
#


class Schema(object):
    """Base class for all Schema classes."""
    _props = None

    def __init__(self, type, other_props=None):
        # Ensure valid ctor args
        if not isinstance(type, basestring):
            fail_msg = 'Schema type must be a string.'
            raise SchemaParseException(fail_msg)
        elif type not in VALID_TYPES:
            fail_msg = '%s is not a valid type.' % type
            raise SchemaParseException(fail_msg)

        # add members
        if self._props is None:
            self._props = {}
        self.set_prop('type', type)
        self.type = type
        self._props.update(other_props or {})

    # Read-only properties dict. Printing schemas
    # creates JSON properties directly from this dict.
    props = property(lambda self: self._props)

    # Read-only property dict. Non-reserved properties
    other_props = property(lambda self: get_other_props(self._props, SCHEMA_RESERVED_PROPS),
                           doc="dictionary of non-reserved properties")

    def check_props(self, other, props):
        """Check that the given props are identical in two schemas.

        @arg other: The other schema to check
        @arg props: An iterable of properties to check
        @return bool: True if all the properties match
        """
        return all(getattr(self, prop) == getattr(other, prop) for prop in props)

    def match(self, writer):
        """Return True if the current schema (as reader) matches the writer schema.

        @arg writer: the writer schema to match against.
        @return bool
        """
        raise NotImplemented("Must be implemented by subclasses")

    # utility functions to manipulate properties dict
    def get_prop(self, key):
        return self._props.get(key)

    def set_prop(self, key, value):
        self._props[key] = value

    def __str__(self):
        return json.dumps(self.to_json())

    def to_json(self, names):
        """
        Converts the schema object into its AVRO specification representation.

        Schema types that have names (records, enums, and fixed) must
        be aware of not re-defining schemas that are already listed
        in the parameter names.
        """
        raise Exception("Must be implemented by subclasses.")


class Name(object):
    """Class to describe Avro name."""

    _full = None

    def __init__(self, name_attr, space_attr, default_space):
        """The fullname is determined in one of the following ways:

        - A name and namespace are both specified. For example, one might use "name": "X", "namespace": "org.foo" to indicate the fullname org.foo.X.
        - A fullname is specified. If the name specified contains a dot,
          then it is assumed to be a fullname, and any namespace also specified is ignored.
          For example, use "name": "org.foo.X" to indicate the fullname org.foo.X.
        - A name only is specified, i.e., a name that contains no dots.
          In this case the namespace is taken from the most tightly enclosing schema or protocol.
          For example, if "name": "X" is specified, and this occurs within a field of
          the record definition of org.foo.Y, then the fullname is org.foo.X.
          If there is no enclosing namespace then the null namespace is used.

        References to previously defined names are as in the latter two cases above:
        if they contain a dot they are a fullname,
        if they do not contain a dot, the namespace is the namespace of the enclosing definition.

        @arg name_attr: name value read in schema or None.
        @arg space_attr: namespace value read in schema or None. The empty string may be used as a namespace to indicate the null namespace.
        @arg default_space: the current default space or None.
        """
        if name_attr is None:
            return
        if name_attr == "":
            raise SchemaParseException('Name must not be the empty string.')

        if '.' in name_attr or space_attr == "" or not (space_attr or default_space):
            # The empty string may be used as a namespace to indicate the null namespace.
            self._full = name_attr
        else:
            self._full = "{!s}.{!s}".format(space_attr or default_space, name_attr)

        self._validate_fullname(self._full)

    def _validate_fullname(self, fullname):
        for name in fullname.split('.'):
            validate_basename(name)

    def __eq__(self, other):
        """Equality of names is defined on the fullname and is case-sensitive."""
        return isinstance(other, Name) and self.fullname == other.fullname

    @property
    def fullname(self):
        return self._full

    @property
    def space(self):
        """Back out a namespace from full name."""
        if self._full is None:
            return None
        return self._full.rsplit(".", 1)[0] if "." in self._full else None

    def get_space(self):
        warnings.warn('Name.get_space() is deprecated in favor of Name.space')
        return self.space


class Names(object):
    """Track name set and default namespace during parsing."""

    def __init__(self, default_namespace=None):
        self.names = {}
        self.default_namespace = default_namespace

    def has_name(self, name_attr, space_attr):
        test = Name(name_attr, space_attr, self.default_namespace).fullname
        return test in self.names

    def get_name(self, name_attr, space_attr):
        test = Name(name_attr, space_attr, self.default_namespace).fullname
        if test not in self.names:
            return None
        return self.names[test]

    def prune_namespace(self, properties):
        """given a properties, return properties with namespace removed if
        it matches the own default namespace"""
        if self.default_namespace is None:
            # I have no default -- no change
            return properties
        if 'namespace' not in properties:
            # he has no namespace - no change
            return properties
        if properties['namespace'] != self.default_namespace:
            # we're different - leave his stuff alone
            return properties
        # we each have a namespace and it's redundant. delete his.
        prunable = properties.copy()
        del(prunable['namespace'])
        return prunable

    def add_name(self, name_attr, space_attr, new_schema):
        """
        Add a new schema object to the name set.

          @arg name_attr: name value read in schema
          @arg space_attr: namespace value read in schema.

          @return: the Name that was just added.
        """
        to_add = Name(name_attr, space_attr, self.default_namespace)

        if to_add.fullname in VALID_TYPES:
            fail_msg = '%s is a reserved type name.' % to_add.fullname
            raise SchemaParseException(fail_msg)
        elif to_add.fullname in self.names:
            fail_msg = 'The name "%s" is already in use.' % to_add.fullname
            raise SchemaParseException(fail_msg)

        self.names[to_add.fullname] = new_schema
        return to_add


class NamedSchema(Schema):
    """Named Schemas specified in NAMED_TYPES."""

    def __init__(self, type, name, namespace=None, names=None, other_props=None):
        # Ensure valid ctor args
        if not name:
            fail_msg = 'Named Schemas must have a non-empty name.'
            raise SchemaParseException(fail_msg)
        elif not isinstance(name, basestring):
            fail_msg = 'The name property must be a string.'
            raise SchemaParseException(fail_msg)
        elif namespace is not None and not isinstance(namespace, basestring):
            fail_msg = 'The namespace property must be a string.'
            raise SchemaParseException(fail_msg)

        # Call parent ctor
        Schema.__init__(self, type, other_props)

        # Add class members
        new_name = names.add_name(name, namespace, self)

        # Store name and namespace as they were read in origin schema
        self.set_prop('name', name)
        if namespace is not None:
            self.set_prop('namespace', new_name.space)

        # Store full name as calculated from name, namespace
        self._fullname = new_name.fullname

    def name_ref(self, names):
        return self.name if self.namespace == names.default_namespace else self.fullname

    # read-only properties
    name = property(lambda self: self.get_prop('name'))
    namespace = property(lambda self: self.get_prop('namespace'))
    fullname = property(lambda self: self._fullname)

#
# Logical type class
#


class LogicalSchema(object):
    def __init__(self, logical_type):
        self.logical_type = logical_type

#
# Decimal logical schema
#


class DecimalLogicalSchema(LogicalSchema):
    def __init__(self, precision, scale=0, max_precision=0):
        if not isinstance(precision, int) or precision <= 0:
            raise IgnoredLogicalType(
                "Invalid decimal precision {}. Must be a positive integer.".format(precision))

        if precision > max_precision:
            raise IgnoredLogicalType(
                "Invalid decimal precision {}. Max is {}.".format(precision, max_precision))

        if not isinstance(scale, int) or scale < 0:
            raise IgnoredLogicalType(
                "Invalid decimal scale {}. Must be a positive integer.".format(scale))

        if scale > precision:
            raise IgnoredLogicalType("Invalid decimal scale {}. Cannot be greater than precision {}."
                                     .format(scale, precision))

        super(DecimalLogicalSchema, self).__init__('decimal')


class Field(object):
    def __init__(self, type, name, has_default, default=None,
                 order=None, names=None, doc=None, other_props=None):
        # Ensure valid ctor args
        if not name:
            fail_msg = 'Fields must have a non-empty name.'
            raise SchemaParseException(fail_msg)
        elif not isinstance(name, basestring):
            fail_msg = 'The name property must be a string.'
            raise SchemaParseException(fail_msg)
        elif order is not None and order not in VALID_FIELD_SORT_ORDERS:
            fail_msg = 'The order property %s is not valid.' % order
            raise SchemaParseException(fail_msg)

        # add members
        self._props = {}
        self._has_default = has_default
        self._props.update(other_props or {})

        if (isinstance(type, basestring) and names is not None and
                names.has_name(type, None)):
            type_schema = names.get_name(type, None)
        else:
            try:
                type_schema = make_avsc_object(type, names)
            except Exception as e:
                fail_msg = 'Type property "%s" not a valid Avro schema: %s' % (type, e)
                raise SchemaParseException(fail_msg)
        self.set_prop('type', type_schema)
        self.set_prop('name', name)
        self.type = type_schema
        self.name = name
        # TODO(hammer): check to ensure default is valid
        if has_default:
            self.set_prop('default', default)
        if order is not None:
            self.set_prop('order', order)
        if doc is not None:
            self.set_prop('doc', doc)

    # read-only properties
    default = property(lambda self: self.get_prop('default'))
    has_default = property(lambda self: self._has_default)
    order = property(lambda self: self.get_prop('order'))
    doc = property(lambda self: self.get_prop('doc'))
    props = property(lambda self: self._props)

    # Read-only property dict. Non-reserved properties
    other_props = property(lambda self: get_other_props(self._props, FIELD_RESERVED_PROPS),
                           doc="dictionary of non-reserved properties")

# utility functions to manipulate properties dict
    def get_prop(self, key):
        return self._props.get(key)

    def set_prop(self, key, value):
        self._props[key] = value

    def __str__(self):
        return json.dumps(self.to_json())

    def to_json(self, names=None):
        if names is None:
            names = Names()
        to_dump = self.props.copy()
        to_dump['type'] = self.type.to_json(names)
        return to_dump

    def __eq__(self, that):
        to_cmp = json.loads(str(self))
        return to_cmp == json.loads(str(that))

#
# Primitive Types
#


class PrimitiveSchema(Schema):
    """Valid primitive types are in PRIMITIVE_TYPES."""

    def __init__(self, type, other_props=None):
        # Ensure valid ctor args
        if type not in PRIMITIVE_TYPES:
            raise AvroException("%s is not a valid primitive type." % type)

        # Call parent ctor
        Schema.__init__(self, type, other_props=other_props)

        self.fullname = type

    def match(self, writer):
        """Return True if the current schema (as reader) matches the writer schema.

        @arg writer: the schema to match against
        @return bool
        """
        return self.type == writer.type or {
            'float': self.type == 'double',
            'int': self.type in {'double', 'float', 'long'},
            'long': self.type in {'double', 'float', },
        }.get(writer.type, False)

    def to_json(self, names=None):
        if len(self.props) == 1:
            return self.fullname
        else:
            return self.props

    def __eq__(self, that):
        return self.props == that.props

#
# Decimal Bytes Type
#


class BytesDecimalSchema(PrimitiveSchema, DecimalLogicalSchema):
    def __init__(self, precision, scale=0, other_props=None):
        DecimalLogicalSchema.__init__(self, precision, scale, max_precision=((1 << 31) - 1))
        PrimitiveSchema.__init__(self, 'bytes', other_props)
        self.set_prop('precision', precision)
        self.set_prop('scale', scale)

    # read-only properties
    precision = property(lambda self: self.get_prop('precision'))
    scale = property(lambda self: self.get_prop('scale'))

    def to_json(self, names=None):
        return self.props

    def __eq__(self, that):
        return self.props == that.props


#
# Complex Types (non-recursive)
#
class FixedSchema(NamedSchema):
    def __init__(self, name, namespace, size, names=None, other_props=None):
        # Ensure valid ctor args
        if not isinstance(size, int) or size < 0:
            fail_msg = 'Fixed Schema requires a valid positive integer for size property.'
            raise AvroException(fail_msg)

        # Call parent ctor
        NamedSchema.__init__(self, 'fixed', name, namespace, names, other_props)

        # Add class members
        self.set_prop('size', size)

    # read-only properties
    size = property(lambda self: self.get_prop('size'))

    def match(self, writer):
        """Return True if the current schema (as reader) matches the writer schema.

        @arg writer: the schema to match against
        @return bool
        """
        return self.type == writer.type and self.check_props(writer, ['fullname', 'size'])

    def to_json(self, names=None):
        if names is None:
            names = Names()
        if self.fullname in names.names:
            return self.name_ref(names)
        else:
            names.names[self.fullname] = self
            return names.prune_namespace(self.props)

    def __eq__(self, that):
        return self.props == that.props

#
# Decimal Fixed Type
#


class FixedDecimalSchema(FixedSchema, DecimalLogicalSchema):
    def __init__(self, size, name, precision, scale=0, namespace=None, names=None, other_props=None):
        max_precision = int(math.floor(math.log10(2) * (8 * size - 1)))
        DecimalLogicalSchema.__init__(self, precision, scale, max_precision)
        FixedSchema.__init__(self, name, namespace, size, names, other_props)
        self.set_prop('precision', precision)
        self.set_prop('scale', scale)

    # read-only properties
    precision = property(lambda self: self.get_prop('precision'))
    scale = property(lambda self: self.get_prop('scale'))

    def to_json(self, names=None):
        return self.props

    def __eq__(self, that):
        return self.props == that.props


class EnumSchema(NamedSchema):
    def __init__(self, name, namespace, symbols, names=None, doc=None, other_props=None, validate_enum_symbols=True):
        """
        @arg validate_enum_symbols: If False, will allow enum symbols that are not valid Avro names.
        """
        if validate_enum_symbols:
            for symbol in symbols:
                try:
                    validate_basename(symbol)
                except InvalidName:
                    raise InvalidName("An enum symbol must be a valid schema name.")

        if len(set(symbols)) < len(symbols):
            fail_msg = 'Duplicate symbol: %s' % symbols
            raise AvroException(fail_msg)

        # Call parent ctor
        NamedSchema.__init__(self, 'enum', name, namespace, names, other_props)

        # Add class members
        self.set_prop('symbols', symbols)
        if doc is not None:
            self.set_prop('doc', doc)

    # read-only properties
    symbols = property(lambda self: self.get_prop('symbols'))
    doc = property(lambda self: self.get_prop('doc'))

    def match(self, writer):
        """Return True if the current schema (as reader) matches the writer schema.

        @arg writer: the schema to match against
        @return bool
        """
        return self.type == writer.type and self.check_props(writer, ['fullname'])

    def to_json(self, names=None):
        if names is None:
            names = Names()
        if self.fullname in names.names:
            return self.name_ref(names)
        else:
            names.names[self.fullname] = self
            return names.prune_namespace(self.props)

    def __eq__(self, that):
        return self.props == that.props

#
# Complex Types (recursive)
#


class ArraySchema(Schema):
    def __init__(self, items, names=None, other_props=None):
        # Call parent ctor
        Schema.__init__(self, 'array', other_props)
        # Add class members

        if isinstance(items, basestring) and names.has_name(items, None):
            items_schema = names.get_name(items, None)
        else:
            try:
                items_schema = make_avsc_object(items, names)
            except SchemaParseException as e:
                fail_msg = 'Items schema (%s) not a valid Avro schema: %s (known names: %s)' % (items, e, names.names.keys())
                raise SchemaParseException(fail_msg)

        self.set_prop('items', items_schema)

    # read-only properties
    items = property(lambda self: self.get_prop('items'))

    def match(self, writer):
        """Return True if the current schema (as reader) matches the writer schema.

        @arg writer: the schema to match against
        @return bool
        """
        return self.type == writer.type and self.items.check_props(writer.items, ['type'])

    def to_json(self, names=None):
        if names is None:
            names = Names()
        to_dump = self.props.copy()
        item_schema = self.get_prop('items')
        to_dump['items'] = item_schema.to_json(names)
        return to_dump

    def __eq__(self, that):
        to_cmp = json.loads(str(self))
        return to_cmp == json.loads(str(that))


class MapSchema(Schema):
    def __init__(self, values, names=None, other_props=None):
        # Call parent ctor
        Schema.__init__(self, 'map', other_props)

        # Add class members
        if isinstance(values, basestring) and names.has_name(values, None):
            values_schema = names.get_name(values, None)
        else:
            try:
                values_schema = make_avsc_object(values, names)
            except SchemaParseException:
                raise
            except Exception:
                raise SchemaParseException('Values schema is not a valid Avro schema.')

        self.set_prop('values', values_schema)

    # read-only properties
    values = property(lambda self: self.get_prop('values'))

    def match(self, writer):
        """Return True if the current schema (as reader) matches the writer schema.

        @arg writer: the schema to match against
        @return bool
        """
        return writer.type == self.type and self.values.check_props(writer.values, ['type'])

    def to_json(self, names=None):
        if names is None:
            names = Names()
        to_dump = self.props.copy()
        to_dump['values'] = self.get_prop('values').to_json(names)
        return to_dump

    def __eq__(self, that):
        to_cmp = json.loads(str(self))
        return to_cmp == json.loads(str(that))


class UnionSchema(Schema):
    """
    names is a dictionary of schema objects
    """

    def __init__(self, schemas, names=None):
        # Ensure valid ctor args
        if not isinstance(schemas, list):
            fail_msg = 'Union schema requires a list of schemas.'
            raise SchemaParseException(fail_msg)

        # Call parent ctor
        Schema.__init__(self, 'union')

        # Add class members
        schema_objects = []
        for schema in schemas:
            if isinstance(schema, basestring) and names.has_name(schema, None):
                new_schema = names.get_name(schema, None)
            else:
                try:
                    new_schema = make_avsc_object(schema, names)
                except Exception as e:
                    raise SchemaParseException('Union item must be a valid Avro schema: %s' % str(e))
            # check the new schema
            if (new_schema.type in VALID_TYPES and new_schema.type not in NAMED_TYPES and
                    new_schema.type in [schema.type for schema in schema_objects]):
                raise SchemaParseException('%s type already in Union' % new_schema.type)
            elif new_schema.type == 'union':
                raise SchemaParseException('Unions cannot contain other unions.')
            else:
                schema_objects.append(new_schema)
        self._schemas = schema_objects

    # read-only properties
    schemas = property(lambda self: self._schemas)

    def match(self, writer):
        """Return True if the current schema (as reader) matches the writer schema.

        @arg writer: the schema to match against
        @return bool
        """
        return writer.type in {'union', 'error_union'} or any(s.match(writer) for s in self.schemas)

    def to_json(self, names=None):
        if names is None:
            names = Names()
        to_dump = []
        for schema in self.schemas:
            to_dump.append(schema.to_json(names))
        return to_dump

    def __eq__(self, that):
        to_cmp = json.loads(str(self))
        return to_cmp == json.loads(str(that))


class ErrorUnionSchema(UnionSchema):
    def __init__(self, schemas, names=None):
        # Prepend "string" to handle system errors
        UnionSchema.__init__(self, ['string'] + schemas, names)

    def to_json(self, names=None):
        if names is None:
            names = Names()
        to_dump = []
        for schema in self.schemas:
            # Don't print the system error schema
            if schema.type == 'string':
                continue
            to_dump.append(schema.to_json(names))
        return to_dump


class RecordSchema(NamedSchema):
    @staticmethod
    def make_field_objects(field_data, names):
        """We're going to need to make message parameters too."""
        field_objects = []
        field_names = []
        for i, field in enumerate(field_data):
            if callable(getattr(field, 'get', None)):
                type = field.get('type')
                name = field.get('name')

                # null values can have a default value of None
                has_default = False
                default = None
                if 'default' in field:
                    has_default = True
                    default = field.get('default')

                order = field.get('order')
                doc = field.get('doc')
                other_props = get_other_props(field, FIELD_RESERVED_PROPS)
                new_field = Field(type, name, has_default, default, order, names, doc,
                                  other_props)
                # make sure field name has not been used yet
                if new_field.name in field_names:
                    fail_msg = 'Field name %s already in use.' % new_field.name
                    raise SchemaParseException(fail_msg)
                field_names.append(new_field.name)
            else:
                raise SchemaParseException('Not a valid field: %s' % field)
            field_objects.append(new_field)
        return field_objects

    def match(self, writer):
        """Return True if the current schema (as reader) matches the other schema.

        @arg writer: the schema to match against
        @return bool
        """
        return writer.type == self.type and (self.type == 'request' or self.check_props(writer, ['fullname']))

    def __init__(self, name, namespace, fields, names=None, schema_type='record',
                 doc=None, other_props=None):
        # Ensure valid ctor args
        if fields is None:
            fail_msg = 'Record schema requires a non-empty fields property.'
            raise SchemaParseException(fail_msg)
        elif not isinstance(fields, list):
            fail_msg = 'Fields property must be a list of Avro schemas.'
            raise SchemaParseException(fail_msg)

        # Call parent ctor (adds own name to namespace, too)
        if schema_type == 'request':
            Schema.__init__(self, schema_type, other_props)
        else:
            NamedSchema.__init__(self, schema_type, name, namespace, names,
                                 other_props)

        if schema_type == 'record':
            old_default = names.default_namespace
            names.default_namespace = Name(name, namespace,
                                           names.default_namespace).space

        # Add class members
        field_objects = RecordSchema.make_field_objects(fields, names)
        self.set_prop('fields', field_objects)
        if doc is not None:
            self.set_prop('doc', doc)

        if schema_type == 'record':
            names.default_namespace = old_default

    # read-only properties
    fields = property(lambda self: self.get_prop('fields'))
    doc = property(lambda self: self.get_prop('doc'))

    @property
    def fields_dict(self):
        fields_dict = {}
        for field in self.fields:
            fields_dict[field.name] = field
        return fields_dict

    def to_json(self, names=None):
        if names is None:
            names = Names()
        # Request records don't have names
        if self.type == 'request':
            return [f.to_json(names) for f in self.fields]

        if self.fullname in names.names:
            return self.name_ref(names)
        else:
            names.names[self.fullname] = self

        to_dump = names.prune_namespace(self.props.copy())
        to_dump['fields'] = [f.to_json(names) for f in self.fields]
        return to_dump

    def __eq__(self, that):
        to_cmp = json.loads(str(self))
        return to_cmp == json.loads(str(that))


#
# Date Type
#

class DateSchema(LogicalSchema, PrimitiveSchema):
    def __init__(self, other_props=None):
        LogicalSchema.__init__(self, constants.DATE)
        PrimitiveSchema.__init__(self, 'int', other_props)

    def to_json(self, names=None):
        return self.props

    def __eq__(self, that):
        return self.props == that.props

#
# time-millis Type
#


class TimeMillisSchema(LogicalSchema, PrimitiveSchema):
    def __init__(self, other_props=None):
        LogicalSchema.__init__(self, constants.TIME_MILLIS)
        PrimitiveSchema.__init__(self, 'int', other_props)

    def to_json(self, names=None):
        return self.props

    def __eq__(self, that):
        return self.props == that.props

#
# time-micros Type
#


class TimeMicrosSchema(LogicalSchema, PrimitiveSchema):
    def __init__(self, other_props=None):
        LogicalSchema.__init__(self, constants.TIME_MICROS)
        PrimitiveSchema.__init__(self, 'long', other_props)

    def to_json(self, names=None):
        return self.props

    def __eq__(self, that):
        return self.props == that.props

#
# timestamp-millis Type
#


class TimestampMillisSchema(LogicalSchema, PrimitiveSchema):
    def __init__(self, other_props=None):
        LogicalSchema.__init__(self, constants.TIMESTAMP_MILLIS)
        PrimitiveSchema.__init__(self, 'long', other_props)

    def to_json(self, names=None):
        return self.props

    def __eq__(self, that):
        return self.props == that.props

#
# timestamp-micros Type
#


class TimestampMicrosSchema(LogicalSchema, PrimitiveSchema):
    def __init__(self, other_props=None):
        LogicalSchema.__init__(self, constants.TIMESTAMP_MICROS)
        PrimitiveSchema.__init__(self, 'long', other_props)

    def to_json(self, names=None):
        return self.props

    def __eq__(self, that):
        return self.props == that.props

#
# Module Methods
#


def get_other_props(all_props, reserved_props):
    """
    Retrieve the non-reserved properties from a dictionary of properties
    @args reserved_props: The set of reserved properties to exclude
    """
    if callable(getattr(all_props, 'items', None)):
        return {k: v for k, v in all_props.items() if k not in reserved_props}


def make_bytes_decimal_schema(other_props):
    """Make a BytesDecimalSchema from just other_props."""
    return BytesDecimalSchema(other_props.get('precision'), other_props.get('scale', 0))


def make_logical_schema(logical_type, type_, other_props):
    """Map the logical types to the appropriate literal type and schema class."""
    logical_types = {
        (constants.DATE, 'int'): DateSchema,
        (constants.DECIMAL, 'bytes'): make_bytes_decimal_schema,
        # The fixed decimal schema is handled later by returning None now.
        (constants.DECIMAL, 'fixed'): lambda x: None,
        (constants.TIMESTAMP_MICROS, 'long'): TimestampMicrosSchema,
        (constants.TIMESTAMP_MILLIS, 'long'): TimestampMillisSchema,
        (constants.TIME_MICROS, 'long'): TimeMicrosSchema,
        (constants.TIME_MILLIS, 'int'): TimeMillisSchema,
    }
    try:
        schema_type = logical_types.get((logical_type, type_), None)
        if schema_type is not None:
            return schema_type(other_props)

        expected_types = sorted(literal_type for lt, literal_type in logical_types if lt == logical_type)
        if expected_types:
            warnings.warn(
                IgnoredLogicalType("Logical type {} requires literal type {}, not {}.".format(
                    logical_type, "/".join(expected_types), type_)))
        else:
            warnings.warn(IgnoredLogicalType("Unknown {}, using {}.".format(logical_type, type_)))
    except IgnoredLogicalType as warning:
        warnings.warn(warning)
    return None


def make_avsc_object(json_data, names=None, validate_enum_symbols=True):
    """
    Build Avro Schema from data parsed out of JSON string.

    @arg names: A Names object (tracks seen names and default space)
    @arg validate_enum_symbols: If False, will allow enum symbols that are not valid Avro names.
    """
    if names is None:
        names = Names()

    # JSON object (non-union)
    if callable(getattr(json_data, 'get', None)):
        type = json_data.get('type')
        other_props = get_other_props(json_data, SCHEMA_RESERVED_PROPS)
        logical_type = json_data.get('logicalType')
        if logical_type:
            logical_schema = make_logical_schema(logical_type, type, other_props or {})
            if logical_schema is not None:
                return logical_schema
        if type in NAMED_TYPES:
            name = json_data.get('name')
            namespace = json_data.get('namespace', names.default_namespace)
            if type == 'fixed':
                size = json_data.get('size')
                if logical_type == 'decimal':
                    precision = json_data.get('precision')
                    scale = 0 if json_data.get('scale') is None else json_data.get('scale')
                    try:
                        return FixedDecimalSchema(size, name, precision, scale, namespace, names, other_props)
                    except IgnoredLogicalType as warning:
                        warnings.warn(warning)
                return FixedSchema(name, namespace, size, names, other_props)
            elif type == 'enum':
                symbols = json_data.get('symbols')
                doc = json_data.get('doc')
                return EnumSchema(name, namespace, symbols, names, doc, other_props, validate_enum_symbols)
            elif type in ['record', 'error']:
                fields = json_data.get('fields')
                doc = json_data.get('doc')
                return RecordSchema(name, namespace, fields, names, type, doc, other_props)
            else:
                raise SchemaParseException('Unknown Named Type: %s' % type)
        if type in PRIMITIVE_TYPES:
            return PrimitiveSchema(type, other_props)
        if type in VALID_TYPES:
            if type == 'array':
                items = json_data.get('items')
                return ArraySchema(items, names, other_props)
            elif type == 'map':
                values = json_data.get('values')
                return MapSchema(values, names, other_props)
            elif type == 'error_union':
                declared_errors = json_data.get('declared_errors')
                return ErrorUnionSchema(declared_errors, names)
            else:
                raise SchemaParseException('Unknown Valid Type: %s' % type)
        elif type is None:
            raise SchemaParseException('No "type" property: %s' % json_data)
        else:
            raise SchemaParseException('Undefined type: %s' % type)
    # JSON array (union)
    elif isinstance(json_data, list):
        return UnionSchema(json_data, names)
    # JSON string (primitive)
    elif json_data in PRIMITIVE_TYPES:
        return PrimitiveSchema(json_data)
    # not for us!
    else:
        fail_msg = "Could not make an Avro Schema object from %s." % json_data
        raise SchemaParseException(fail_msg)

# TODO(hammer): make method for reading from a file?


def parse(json_string, validate_enum_symbols=True):
    """Constructs the Schema from the JSON text.

    @arg validate_enum_symbols: If False, will allow enum symbols that are not valid Avro names.
    """
    # parse the JSON
    try:
        json_data = json.loads(json_string)
    except Exception as e:
        msg = 'Error parsing JSON: {}, error = {}'.format(json_string, e)
        new_exception = SchemaParseException(msg)
        traceback = sys.exc_info()[2]
        if not hasattr(new_exception, 'with_traceback'):
            raise (new_exception, None, traceback)  # Python 2 syntax
        raise new_exception.with_traceback(traceback)

    # Initialize the names object
    names = Names()

    # construct the Avro Schema object
    return make_avsc_object(json_data, names, validate_enum_symbols)



#!/usr/bin/env python

##
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
# https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Read/Write Avro File Object Containers."""

from __future__ import absolute_import, division, print_function

import io
import os
import random
import zlib

import avro.io
import avro.schema
from avro.codecs import Codecs

#
# Constants
#
VERSION = 1
MAGIC = bytes(b'Obj' + bytearray([VERSION]))
MAGIC_SIZE = len(MAGIC)
SYNC_SIZE = 16
SYNC_INTERVAL = 4000 * SYNC_SIZE  # TODO(hammer): make configurable
META_SCHEMA = avro.schema.parse("""\
{"type": "record", "name": "org.apache.avro.file.Header",
 "fields" : [
   {"name": "magic", "type": {"type": "fixed", "name": "magic", "size": %d}},
   {"name": "meta", "type": {"type": "map", "values": "bytes"}},
   {"name": "sync", "type": {"type": "fixed", "name": "sync", "size": %d}}]}
""" % (MAGIC_SIZE, SYNC_SIZE))

NULL_CODEC = 'null'
VALID_CODECS = Codecs.supported_codec_names()
VALID_ENCODINGS = ['binary']  # not used yet

CODEC_KEY = "avro.codec"
SCHEMA_KEY = "avro.schema"

#
# Exceptions
#


class DataFileException(avro.schema.AvroException):
    """
    Raised when there's a problem reading or writing file object containers.
    """

    def __init__(self, fail_msg):
        avro.schema.AvroException.__init__(self, fail_msg)

#
# Write Path
#


class _DataFile(object):
    """Mixin for methods common to both reading and writing."""

    block_count = 0
    _meta = None
    _sync_marker = None

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        # Perform a close if there's no exception
        if type is None:
            self.close()

    def get_meta(self, key):
        return self.meta.get(key)

    def set_meta(self, key, val):
        self.meta[key] = val

    @property
    def sync_marker(self):
        return self._sync_marker

    @property
    def meta(self):
        """Read-only dictionary of metadata for this datafile."""
        if self._meta is None:
            self._meta = {}
        return self._meta

    @property
    def codec(self):
        """Meta are stored as bytes, but codec is returned as a string."""
        try:
            return self.get_meta(CODEC_KEY).decode()
        except AttributeError:
            return "null"

    @codec.setter
    def codec(self, value):
        """Meta are stored as bytes, but codec is set as a string."""
        if value not in VALID_CODECS:
            raise DataFileException("Unknown codec: {!r}".format(value))
        self.set_meta(CODEC_KEY, value.encode())

    @property
    def schema(self):
        """Meta are stored as bytes, but schema is returned as a string."""
        return self.get_meta(SCHEMA_KEY).decode()

    @schema.setter
    def schema(self, value):
        """Meta are stored as bytes, but schema is set as a string."""
        self.set_meta(SCHEMA_KEY, value.encode())


class DataFileWriter(_DataFile):

    # TODO(hammer): make 'encoder' a metadata property
    def __init__(self, writer, datum_writer, writers_schema=None, codec=NULL_CODEC):
        """
        If the schema is not present, presume we're appending.

        @param writer: File-like object to write into.
        """
        self._writer = writer
        self._encoder = avro.io.BinaryEncoder(writer)
        self._datum_writer = datum_writer
        self._buffer_writer = io.BytesIO()
        self._buffer_encoder = avro.io.BinaryEncoder(self._buffer_writer)
        self.block_count = 0
        self._header_written = False

        if writers_schema is not None:
            self._sync_marker = generate_sixteen_random_bytes()
            self.codec = codec
            self.schema = str(writers_schema)
            self.datum_writer.writers_schema = writers_schema
        else:
            # open writer for reading to collect metadata
            dfr = DataFileReader(writer, avro.io.DatumReader())

            # TODO(hammer): collect arbitrary metadata
            # collect metadata
            self._sync_marker = dfr.sync_marker
            self.codec = dfr.codec

            # get schema used to write existing file
            self.schema = schema_from_file = dfr.schema
            self.datum_writer.writers_schema = avro.schema.parse(schema_from_file)

            # seek to the end of the file and prepare for writing
            writer.seek(0, 2)
            self._header_written = True

    # read-only properties
    writer = property(lambda self: self._writer)
    encoder = property(lambda self: self._encoder)
    datum_writer = property(lambda self: self._datum_writer)
    buffer_writer = property(lambda self: self._buffer_writer)
    buffer_encoder = property(lambda self: self._buffer_encoder)

    def _write_header(self):
        header = {'magic': MAGIC,
                  'meta': self.meta,
                  'sync': self.sync_marker}
        self.datum_writer.write_data(META_SCHEMA, header, self.encoder)
        self._header_written = True

    @property
    def codec(self):
        """Meta are stored as bytes, but codec is returned as a string."""
        return self.get_meta(CODEC_KEY).decode()

    @codec.setter
    def codec(self, value):
        """Meta are stored as bytes, but codec is set as a string."""
        if value not in VALID_CODECS:
            raise DataFileException("Unknown codec: {!r}".format(value))
        self.set_meta(CODEC_KEY, value.encode())

    # TODO(hammer): make a schema for blocks and use datum_writer
    def _write_block(self):
        if not self._header_written:
            self._write_header()

        if self.block_count > 0:
            # write number of items in block
            self.encoder.write_long(self.block_count)

            # write block contents
            uncompressed_data = self.buffer_writer.getvalue()
            codec = Codecs.get_codec(self.codec)
            compressed_data, compressed_data_length = codec.compress(uncompressed_data)

            # Write length of block
            self.encoder.write_long(compressed_data_length)

            # Write block
            self.writer.write(compressed_data)

            # write sync marker
            self.writer.write(self.sync_marker)

            # reset buffer
            self.buffer_writer.truncate(0)
            self.buffer_writer.seek(0)
            self.block_count = 0

    def append(self, datum):
        """Append a datum to the file."""
        self.datum_writer.write(datum, self.buffer_encoder)
        self.block_count += 1

        # if the data to write is larger than the sync interval, write the block
        if self.buffer_writer.tell() >= SYNC_INTERVAL:
            self._write_block()

    def sync(self):
        """
        Return the current position as a value that may be passed to
        DataFileReader.seek(long). Forces the end of the current block,
        emitting a synchronization marker.
        """
        self._write_block()
        return self.writer.tell()

    def flush(self):
        """Flush the current state of the file, including metadata."""
        self._write_block()
        self.writer.flush()

    def close(self):
        """Close the file."""
        self.flush()
        self.writer.close()


class DataFileReader(_DataFile):
    """Read files written by DataFileWriter."""
    # TODO(hammer): allow user to specify expected schema?
    # TODO(hammer): allow user to specify the encoder

    def __init__(self, reader, datum_reader):
        self._reader = reader
        self._raw_decoder = avro.io.BinaryDecoder(reader)
        self._datum_decoder = None  # Maybe reset at every block.
        self._datum_reader = datum_reader

        # read the header: magic, meta, sync
        self._read_header()

        # get file length
        self._file_length = self.determine_file_length()

        # get ready to read
        self.block_count = 0
        self.datum_reader.writers_schema = avro.schema.parse(self.schema)

    def __iter__(self):
        return self

    # read-only properties
    reader = property(lambda self: self._reader)
    raw_decoder = property(lambda self: self._raw_decoder)
    datum_decoder = property(lambda self: self._datum_decoder)
    datum_reader = property(lambda self: self._datum_reader)
    file_length = property(lambda self: self._file_length)

    def determine_file_length(self):
        """
        Get file length and leave file cursor where we found it.
        """
        remember_pos = self.reader.tell()
        self.reader.seek(0, 2)
        file_length = self.reader.tell()
        self.reader.seek(remember_pos)
        return file_length

    def is_EOF(self):
        return self.reader.tell() == self.file_length

    def _read_header(self):
        # seek to the beginning of the file to get magic block
        self.reader.seek(0, 0)

        # read header into a dict
        header = self.datum_reader.read_data(
            META_SCHEMA, META_SCHEMA, self.raw_decoder)

        # check magic number
        if header.get('magic') != MAGIC:
            fail_msg = "Not an Avro data file: %s doesn't match %s."\
                       % (header.get('magic'), MAGIC)
            raise avro.schema.AvroException(fail_msg)

        # set metadata
        self._meta = header['meta']

        # set sync marker
        self._sync_marker = header['sync']

    def _read_block_header(self):
        self.block_count = self.raw_decoder.read_long()
        codec = Codecs.get_codec(self.codec)
        self._datum_decoder = codec.decompress(self.raw_decoder)

    def _skip_sync(self):
        """
        Read the length of the sync marker; if it matches the sync marker,
        return True. Otherwise, seek back to where we started and return False.
        """
        proposed_sync_marker = self.reader.read(SYNC_SIZE)
        if proposed_sync_marker != self.sync_marker:
            self.reader.seek(-SYNC_SIZE, 1)
            return False
        return True

    def __next__(self):
        """Return the next datum in the file."""
        while self.block_count == 0:
            if self.is_EOF() or (self._skip_sync() and self.is_EOF()):
                raise StopIteration
            self._read_block_header()

        datum = self.datum_reader.read(self.datum_decoder)
        self.block_count -= 1
        return datum
    next = __next__

    def close(self):
        """Close this reader."""
        self.reader.close()


def generate_sixteen_random_bytes():
    try:
        return os.urandom(16)
    except NotImplementedError:
        return bytes(random.randrange(256) for i in range(16))



#!/usr/bin/env python

##
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
# https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Support for inter-process calls."""

from __future__ import absolute_import, division, print_function

import io
import os
from struct import Struct

import avro.io
from avro import protocol, schema

try:
    import httplib  # type: ignore
except ImportError:
    import http.client as httplib  # type: ignore

try:
    unicode
except NameError:
    unicode = str


def _load(name):
    dir_path = os.path.dirname(__file__)
    rsrc_path = os.path.join(dir_path, name)
    with open(rsrc_path, 'r') as f:
        return f.read()


HANDSHAKE_REQUEST_SCHEMA_JSON = _load('HandshakeRequest.avsc')
HANDSHAKE_RESPONSE_SCHEMA_JSON = _load('HandshakeResponse.avsc')
HANDSHAKE_REQUEST_SCHEMA = schema.parse(HANDSHAKE_REQUEST_SCHEMA_JSON)
HANDSHAKE_RESPONSE_SCHEMA = schema.parse(HANDSHAKE_RESPONSE_SCHEMA_JSON)

HANDSHAKE_REQUESTOR_WRITER = avro.io.DatumWriter(HANDSHAKE_REQUEST_SCHEMA)
HANDSHAKE_REQUESTOR_READER = avro.io.DatumReader(HANDSHAKE_RESPONSE_SCHEMA)
HANDSHAKE_RESPONDER_WRITER = avro.io.DatumWriter(HANDSHAKE_RESPONSE_SCHEMA)
HANDSHAKE_RESPONDER_READER = avro.io.DatumReader(HANDSHAKE_REQUEST_SCHEMA)

META_SCHEMA = schema.parse('{"type": "map", "values": "bytes"}')
META_WRITER = avro.io.DatumWriter(META_SCHEMA)
META_READER = avro.io.DatumReader(META_SCHEMA)

SYSTEM_ERROR_SCHEMA = schema.parse('["string"]')

# protocol cache
REMOTE_HASHES = {}
REMOTE_PROTOCOLS = {}

BIG_ENDIAN_INT_STRUCT = Struct('!I')
BUFFER_HEADER_LENGTH = 4
BUFFER_SIZE = 8192

#
# Exceptions
#


class AvroRemoteException(schema.AvroException):
    """
    Raised when an error message is sent by an Avro requestor or responder.
    """

    def __init__(self, fail_msg=None):
        schema.AvroException.__init__(self, fail_msg)


class ConnectionClosedException(schema.AvroException):
    pass

#
# Base IPC Classes (Requestor/Responder)
#


class BaseRequestor(object):
    """Base class for the client side of a protocol interaction."""

    def __init__(self, local_protocol, transceiver):
        self._local_protocol = local_protocol
        self._transceiver = transceiver
        self._remote_protocol = None
        self._remote_hash = None
        self._send_protocol = None

    # read-only properties
    local_protocol = property(lambda self: self._local_protocol)
    transceiver = property(lambda self: self._transceiver)

    # read/write properties
    def set_remote_protocol(self, new_remote_protocol):
        self._remote_protocol = new_remote_protocol
        REMOTE_PROTOCOLS[self.transceiver.remote_name] = self.remote_protocol
    remote_protocol = property(lambda self: self._remote_protocol,
                               set_remote_protocol)

    def set_remote_hash(self, new_remote_hash):
        self._remote_hash = new_remote_hash
        REMOTE_HASHES[self.transceiver.remote_name] = self.remote_hash
    remote_hash = property(lambda self: self._remote_hash, set_remote_hash)

    def set_send_protocol(self, new_send_protocol):
        self._send_protocol = new_send_protocol
    send_protocol = property(lambda self: self._send_protocol, set_send_protocol)

    def request(self, message_name, request_datum):
        """
        Writes a request message and reads a response or error message.
        """
        # build handshake and call request
        buffer_writer = io.BytesIO()
        buffer_encoder = avro.io.BinaryEncoder(buffer_writer)
        self.write_handshake_request(buffer_encoder)
        self.write_call_request(message_name, request_datum, buffer_encoder)

        # send the handshake and call request; block until call response
        call_request = buffer_writer.getvalue()
        return self.issue_request(call_request, message_name, request_datum)

    def write_handshake_request(self, encoder):
        local_hash = self.local_protocol.md5
        remote_name = self.transceiver.remote_name
        remote_hash = REMOTE_HASHES.get(remote_name)
        if remote_hash is None:
            remote_hash = local_hash
            self.remote_protocol = self.local_protocol
        request_datum = {}
        request_datum['clientHash'] = local_hash
        request_datum['serverHash'] = remote_hash
        if self.send_protocol:
            request_datum['clientProtocol'] = unicode(self.local_protocol)
        HANDSHAKE_REQUESTOR_WRITER.write(request_datum, encoder)

    def write_call_request(self, message_name, request_datum, encoder):
        """
        The format of a call request is:
          * request metadata, a map with values of type bytes
          * the message name, an Avro string, followed by
          * the message parameters. Parameters are serialized according to
            the message's request declaration.
        """
        # request metadata (not yet implemented)
        request_metadata = {}
        META_WRITER.write(request_metadata, encoder)

        # message name
        message = self.local_protocol.messages.get(message_name)
        if message is None:
            raise schema.AvroException('Unknown message: %s' % message_name)
        encoder.write_utf8(message.name)

        # message parameters
        self.write_request(message.request, request_datum, encoder)

    def write_request(self, request_schema, request_datum, encoder):
        datum_writer = avro.io.DatumWriter(request_schema)
        datum_writer.write(request_datum, encoder)

    def read_handshake_response(self, decoder):
        handshake_response = HANDSHAKE_REQUESTOR_READER.read(decoder)
        match = handshake_response.get('match')
        if match == 'BOTH':
            self.send_protocol = False
            return True
        elif match == 'CLIENT':
            if self.send_protocol:
                raise schema.AvroException('Handshake failure.')
            self.remote_protocol = protocol.parse(
                handshake_response.get('serverProtocol'))
            self.remote_hash = handshake_response.get('serverHash')
            self.send_protocol = False
            return True
        elif match == 'NONE':
            if self.send_protocol:
                raise schema.AvroException('Handshake failure.')
            self.remote_protocol = protocol.parse(
                handshake_response.get('serverProtocol'))
            self.remote_hash = handshake_response.get('serverHash')
            self.send_protocol = True
            return False
        else:
            raise schema.AvroException('Unexpected match: %s' % match)

    def read_call_response(self, message_name, decoder):
        """
        The format of a call response is:
          * response metadata, a map with values of type bytes
          * a one-byte error flag boolean, followed by either:
            o if the error flag is false,
              the message response, serialized per the message's response schema.
            o if the error flag is true,
              the error, serialized per the message's error union schema.
        """
        # response metadata
        response_metadata = META_READER.read(decoder)

        # remote response schema
        remote_message_schema = self.remote_protocol.messages.get(message_name)
        if remote_message_schema is None:
            raise schema.AvroException('Unknown remote message: %s' % message_name)

        # local response schema
        local_message_schema = self.local_protocol.messages.get(message_name)
        if local_message_schema is None:
            raise schema.AvroException('Unknown local message: %s' % message_name)

        # error flag
        if not decoder.read_boolean():
            writers_schema = remote_message_schema.response
            readers_schema = local_message_schema.response
            return self.read_response(writers_schema, readers_schema, decoder)
        else:
            writers_schema = remote_message_schema.errors
            readers_schema = local_message_schema.errors
            raise self.read_error(writers_schema, readers_schema, decoder)

    def read_response(self, writers_schema, readers_schema, decoder):
        datum_reader = avro.io.DatumReader(writers_schema, readers_schema)
        result = datum_reader.read(decoder)
        return result

    def read_error(self, writers_schema, readers_schema, decoder):
        datum_reader = avro.io.DatumReader(writers_schema, readers_schema)
        return AvroRemoteException(datum_reader.read(decoder))


class Requestor(BaseRequestor):

    def issue_request(self, call_request, message_name, request_datum):
        call_response = self.transceiver.transceive(call_request)

        # process the handshake and call response
        buffer_decoder = avro.io.BinaryDecoder(io.BytesIO(call_response))
        call_response_exists = self.read_handshake_response(buffer_decoder)
        if call_response_exists:
            return self.read_call_response(message_name, buffer_decoder)
        return self.request(message_name, request_datum)


class Responder(object):
    """Base class for the server side of a protocol interaction."""

    def __init__(self, local_protocol):
        self._local_protocol = local_protocol
        self._local_hash = self.local_protocol.md5
        self._protocol_cache = {}
        self.set_protocol_cache(self.local_hash, self.local_protocol)

    # read-only properties
    local_protocol = property(lambda self: self._local_protocol)
    local_hash = property(lambda self: self._local_hash)
    protocol_cache = property(lambda self: self._protocol_cache)

    # utility functions to manipulate protocol cache
    def get_protocol_cache(self, hash):
        return self.protocol_cache.get(hash)

    def set_protocol_cache(self, hash, protocol):
        self.protocol_cache[hash] = protocol

    def respond(self, call_request):
        """
        Called by a server to deserialize a request, compute and serialize
        a response or error. Compare to 'handle()' in Thrift.
        """
        buffer_reader = io.BytesIO(call_request)
        buffer_decoder = avro.io.BinaryDecoder(buffer_reader)
        buffer_writer = io.BytesIO()
        buffer_encoder = avro.io.BinaryEncoder(buffer_writer)
        error = None
        response_metadata = {}

        try:
            remote_protocol = self.process_handshake(buffer_decoder, buffer_encoder)
            # handshake failure
            if remote_protocol is None:
                return buffer_writer.getvalue()

            # read request using remote protocol
            request_metadata = META_READER.read(buffer_decoder)
            remote_message_name = buffer_decoder.read_utf8()

            # get remote and local request schemas so we can do
            # schema resolution (one fine day)
            remote_message = remote_protocol.messages.get(remote_message_name)
            if remote_message is None:
                fail_msg = 'Unknown remote message: %s' % remote_message_name
                raise schema.AvroException(fail_msg)
            local_message = self.local_protocol.messages.get(remote_message_name)
            if local_message is None:
                fail_msg = 'Unknown local message: %s' % remote_message_name
                raise schema.AvroException(fail_msg)
            writers_schema = remote_message.request
            readers_schema = local_message.request
            request = self.read_request(writers_schema, readers_schema,
                                        buffer_decoder)

            # perform server logic
            try:
                response = self.invoke(local_message, request)
            except AvroRemoteException as e:
                error = e
            except Exception as e:
                error = AvroRemoteException(unicode(e))

            # write response using local protocol
            META_WRITER.write(response_metadata, buffer_encoder)
            buffer_encoder.write_boolean(error is not None)
            if error is None:
                writers_schema = local_message.response
                self.write_response(writers_schema, response, buffer_encoder)
            else:
                writers_schema = local_message.errors
                self.write_error(writers_schema, error, buffer_encoder)
        except schema.AvroException as e:
            error = AvroRemoteException(unicode(e))
            buffer_encoder = avro.io.BinaryEncoder(io.BytesIO())
            META_WRITER.write(response_metadata, buffer_encoder)
            buffer_encoder.write_boolean(True)
            self.write_error(SYSTEM_ERROR_SCHEMA, error, buffer_encoder)
        return buffer_writer.getvalue()

    def process_handshake(self, decoder, encoder):
        handshake_request = HANDSHAKE_RESPONDER_READER.read(decoder)
        handshake_response = {}

        # determine the remote protocol
        client_hash = handshake_request.get('clientHash')
        client_protocol = handshake_request.get('clientProtocol')
        remote_protocol = self.get_protocol_cache(client_hash)
        if remote_protocol is None and client_protocol is not None:
            remote_protocol = protocol.parse(client_protocol)
            self.set_protocol_cache(client_hash, remote_protocol)

        # evaluate remote's guess of the local protocol
        server_hash = handshake_request.get('serverHash')
        if self.local_hash == server_hash:
            if remote_protocol is None:
                handshake_response['match'] = 'NONE'
            else:
                handshake_response['match'] = 'BOTH'
        else:
            if remote_protocol is None:
                handshake_response['match'] = 'NONE'
            else:
                handshake_response['match'] = 'CLIENT'

        if handshake_response['match'] != 'BOTH':
            handshake_response['serverProtocol'] = unicode(self.local_protocol)
            handshake_response['serverHash'] = self.local_hash

        HANDSHAKE_RESPONDER_WRITER.write(handshake_response, encoder)
        return remote_protocol

    def invoke(self, local_message, request):
        """
        Aactual work done by server: cf. handler in thrift.
        """
        pass

    def read_request(self, writers_schema, readers_schema, decoder):
        datum_reader = avro.io.DatumReader(writers_schema, readers_schema)
        return datum_reader.read(decoder)

    def write_response(self, writers_schema, response_datum, encoder):
        datum_writer = avro.io.DatumWriter(writers_schema)
        datum_writer.write(response_datum, encoder)

    def write_error(self, writers_schema, error_exception, encoder):
        datum_writer = avro.io.DatumWriter(writers_schema)
        datum_writer.write(unicode(error_exception), encoder)

#
# Utility classes
#


class FramedReader(object):
    """Wrapper around a file-like object to read framed data."""

    def __init__(self, reader):
        self._reader = reader

    # read-only properties
    reader = property(lambda self: self._reader)

    def read_framed_message(self):
        message = []
        while True:
            buffer = io.BytesIO()
            buffer_length = self._read_buffer_length()
            if buffer_length == 0:
                return b''.join(message)
            while buffer.tell() < buffer_length:
                chunk = self.reader.read(buffer_length - buffer.tell())
                if chunk == '':
                    raise ConnectionClosedException("Reader read 0 bytes.")
                buffer.write(chunk)
            message.append(buffer.getvalue())

    def _read_buffer_length(self):
        read = self.reader.read(BUFFER_HEADER_LENGTH)
        if read == '':
            raise ConnectionClosedException("Reader read 0 bytes.")
        return BIG_ENDIAN_INT_STRUCT.unpack(read)[0]


class FramedWriter(object):
    """Wrapper around a file-like object to write framed data."""

    def __init__(self, writer):
        self._writer = writer

    # read-only properties
    writer = property(lambda self: self._writer)

    def write_framed_message(self, message):
        message_length = len(message)
        total_bytes_sent = 0
        while message_length - total_bytes_sent > 0:
            if message_length - total_bytes_sent > BUFFER_SIZE:
                buffer_length = BUFFER_SIZE
            else:
                buffer_length = message_length - total_bytes_sent
            self.write_buffer(message[total_bytes_sent:
                                      (total_bytes_sent + buffer_length)])
            total_bytes_sent += buffer_length
        # A message is always terminated by a zero-length buffer.
        self.write_buffer_length(0)

    def write_buffer(self, chunk):
        buffer_length = len(chunk)
        self.write_buffer_length(buffer_length)
        self.writer.write(chunk)

    def write_buffer_length(self, n):
        self.writer.write(BIG_ENDIAN_INT_STRUCT.pack(n))

#
# Transceiver Implementations
#


class HTTPTransceiver(object):
    """
    A simple HTTP-based transceiver implementation.
    Useful for clients but not for servers
    """

    def __init__(self, host, port, req_resource='/'):
        self.req_resource = req_resource
        self.conn = httplib.HTTPConnection(host, port)
        self.conn.connect()
        self.remote_name = self.conn.sock.getsockname()

    def transceive(self, request):
        self.write_framed_message(request)
        result = self.read_framed_message()
        return result

    def read_framed_message(self):
        response = self.conn.getresponse()
        response_reader = FramedReader(response)
        framed_message = response_reader.read_framed_message()
        response.read()    # ensure we're ready for subsequent requests
        return framed_message

    def write_framed_message(self, message):
        req_method = 'POST'
        req_headers = {'Content-Type': 'avro/binary'}

        req_body_buffer = FramedWriter(io.BytesIO())
        req_body_buffer.write_framed_message(message)
        req_body = req_body_buffer.writer.getvalue()

        self.conn.request(req_method, self.req_resource, req_body, req_headers)

    def close(self):
        self.conn.close()

#
# Server Implementations (none yet)
#



#!/usr/bin/env python

##
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import, division, print_function

import string
import sys
import time
from random import choice, randint, sample

import avro.datafile
import avro.io
import avro.schema

types = ["A", "CNAME"]


def rand_name():
    return ''.join(sample(string.ascii_lowercase, 15))


def rand_ip():
    return "%s.%s.%s.%s" % (randint(0, 255), randint(0, 255), randint(0, 255), randint(0, 255))


def write(n):
    schema_s = """
    { "type": "record",
      "name": "Query",
    "fields" : [
        {"name": "query", "type": "string"},
        {"name": "response", "type": "string"},
        {"name": "type", "type": "string", "default": "A"}
    ]}"""
    out = open("datafile.avr", 'w')

    schema = avro.schema.parse(schema_s)
    writer = avro.io.DatumWriter(schema)
    dw = avro.datafile.DataFileWriter(out, writer, schema)  # ,codec='deflate')
    for _ in xrange(n):
        response = rand_ip()
        query = rand_name()
        type = choice(types)
        dw.append({'query': query, 'response': response, 'type': type})

    dw.close()


def read():
    f = open("datafile.avr")
    reader = avro.io.DatumReader()
    af = avro.datafile.DataFileReader(f, reader)

    x = 0
    for _ in af:
        pass


def t(f, *args):
    s = time.time()
    f(*args)
    e = time.time()
    return e - s


if __name__ == "__main__":
    n = int(sys.argv[1])
    print("Write %0.4f" % t(write, n))
    print("Read %0.4f" % t(read))
